# -*- coding: utf-8 -*-
# @Author: Pengyao Ping
# @Date:   2023-02-16 11:01:06
# @Last Modified by:   Pengyao Ping
# @Last Modified time: 2023-05-14 11:00:24
# The codes in this section for features construction are modified from https://github.com/Bonidia/MathFeature, if you re-use these codes or methods, please also cite the original publication of Robson P Bonidia, Douglas S Domingues, Danilo S Sanches, André C P L F de Carvalho, MathFeature: feature extraction package for DNA, RNA and protein sequences based on mathematical descriptors, Briefings in Bioinformatics, 2021; bbab434, https://doi.org/10.1093/bib/bbab434.


import numpy as np
import statistics
from scipy.fftpack import fft
import math

class FourierTransform():
    def __init__(self, read):
        self.read = read.upper()

    def feature_extraction(self, spectrum, spectrumTwo):
        features = []
        average = sum(spectrum)/len(spectrum)
        features.append(average)
        ###################################
        median = np.median(spectrum)
        features.append(median)
        ###################################
        maximum = np.max(spectrum)
        features.append(maximum)
        ###################################
        minimum = np.min(spectrum)
        features.append(minimum)
        ###################################
        if average != 0:
            peak = (len(spectrum)/3)/(average)
        else:
            peak = 0
        features.append(peak)
        ###################################
        if np.mean(spectrumTwo) != 0:
            peak_two = (len(spectrumTwo)/3)/(np.mean(spectrumTwo))
        else:
            peak_two = 0
        features.append(peak_two)
        ###################################
        standard_deviation = np.std(spectrum) # standard deviation
        features.append(standard_deviation)
        ###################################
        if len(spectrum) >= 2:
            standard_deviation_pop = statistics.stdev(spectrum) # population sample standard deviation 
        else:
            standard_deviation_pop = 0
        features.append(standard_deviation_pop)
        ###################################
        percentile15 = np.percentile(spectrum, 15)
        features.append(percentile15)
        ###################################
        percentile25 = np.percentile(spectrum, 25)
        features.append(percentile25)
        ###################################
        percentile50 = np.percentile(spectrum, 50)
        features.append(percentile50)
        ###################################
        percentile75 = np.percentile(spectrum, 75)
        features.append(percentile75)
        ###################################
        amplitude = maximum - minimum
        features.append(amplitude)
        ###################################
        # mode = statistics.mode(spectrum)
        ###################################
        if len(spectrum) >= 2:
            variance = statistics.variance(spectrum)
        else:
            variance = 0
        features.append(variance)
        ###################################
        interquartile_range = np.percentile(spectrum, 75) - np.percentile(spectrum, 25)
        features.append(interquartile_range)
        ###################################
        semi_interquartile_range = (np.percentile(spectrum, 75) - np.percentile(spectrum, 25))/2 
        features.append(semi_interquartile_range)
        ###################################
        if average != 0:
            coefficient_of_variation = standard_deviation/average
        else:
            coefficient_of_variation = 0
        features.append(coefficient_of_variation)
        ###################################
        if standard_deviation != 0:
            skewness = (3 * (average - median))/standard_deviation
        else:
            skewness = 0
        features.append(skewness)   
        ###################################
        if (2 * (np.percentile(spectrum, 90) - np.percentile(spectrum, 10))) == 0:
            kurtosis = 0
        else:
            kurtosis = (np.percentile(spectrum, 75) - np.percentile(spectrum, 25)) / (2 * (np.percentile(spectrum, 90) - np.percentile(spectrum, 10))) 
        # print(kurtosis)
        features.append(kurtosis)
        ###################################
        return features

    def binary_fourier(self):
        # if len(self.read) > 0:
        spectrum = []
        spectrumTwo = []
        A = []
        C = []
        T = []
        G = []
        for nucle in self.read:
            if nucle == "A":
                A.append(1)
            else:
                A.append(0)
            if nucle == "C":
                C.append(1)
            else:
                C.append(0)
            if nucle == "T" or nucle =="U":
                T.append(1)
            else:
                T.append(0)
            if nucle == "G":
                G.append(1)
            else:
                G.append(0)
        FA = fft(A)
        FC = fft(C)
        FT = fft(T)
        FG = fft(G)
        for i in range(len(self.read)):
            specTotal = (abs(FA[i])**2) + (abs(FC[i])**2) + (abs(FT[i])**2) + (abs(FG[i])**2)
            specTwo = (abs(FA[i])) + (abs(FC[i])) + (abs(FT[i])) + (abs(FG[i]))
            spectrum.append(specTotal)
            spectrumTwo.append(specTwo)
        features = self.feature_extraction(spectrum, spectrumTwo)
        return features

    def zcurve_fourier(self):
        # if len(self.read) > 0:
        spectrum = []
        spectrumTwo = []
        ###################################
        ###################################
        R = 0 # x[n] = (An + Gn) − (Cn + Tn) ≡ Rn − Yn
        Y = 0
        M = 0 # y[n] = (An + Cn) − (Gn + Tn) ≡ Mn − Kn
        K = 0
        W = 0 # z[n] = (An + Tn) − (Cn + Gn) ≡ Wn − Sn
        S = 0
        ###################################
        ###################################
        x = []
        y = []
        z = []
        for nucle in self.read:
            if nucle == "A" or nucle == "G":
                R += 1
                x.append((R)-(Y))
            else:
                Y += 1
                x.append((R)-(Y))
            if nucle == "A" or nucle == "C":
                M += 1
                y.append((M)-(K))
            else:
                K += 1
                y.append((M)-(K))
            if nucle == "A" or nucle == "T" or nucle == "U":
                W += 1
                z.append((W)-(S))
            else:
                S += 1
                z.append((W)-(S))
        # if len(x) == 0:
        #     x = [0]
        # if len(y) == 0:
        #     y = [0]
        # if len(z) == 0:
        #     z = [0]

        FX = fft(x)
        FY = fft(y)
        FZ = fft(z)
        for i in range(len(self.read)):
            specTotal = (abs(FX[i])**2) + (abs(FY[i])**2) + (abs(FZ[i])**2)
            specTwo = (abs(FX[i])) + (abs(FY[i])) + (abs(FZ[i]))
            spectrum.append(specTotal)
            spectrumTwo.append(specTwo)
        features = self.feature_extraction(spectrum, spectrumTwo)
        return features

    def integer_fourier(self):
        # if len(self.read) > 0:
        spectrum = []
        spectrumTwo = []
        integer = []
        for nucle in self.read:
            if nucle == "T" or nucle == "U":
                integer.append(0)
            elif nucle == "C":
                integer.append(1)
            elif nucle == "A":
                integer.append(2)
            else:
                integer.append(3)
        # if len(integer) == 0:
        #     integer = [0]
        FI = fft(integer)
        for i in range(len(self.read)):
            specTotal = (abs(FI[i])**2)
            specTwo = (abs(FI[i]))
            spectrum.append(specTotal)
            spectrumTwo.append(specTwo)
        features = self.feature_extraction(spectrum, spectrumTwo)
        return features

    def real_fourier(self):
        # if len(self.read) > 0:
        spectrum = []
        spectrumTwo = []
        real = []
        for nucle in self.read:
            if nucle == "T" or nucle == "U":
                real.append(1.5)
            elif nucle == "C":
                real.append(0.5)
            elif nucle == "A":
                real.append(-1.5)
            else:
                real.append(-0.5)
        # if len(real) == 0:
        #     real = [0]
        FR = fft(real)
        for i in range(len(self.read)):
            specTotal = (abs(FR[i])**2)
            specTwo = (abs(FR[i]))
            spectrum.append(specTotal)
            spectrumTwo.append(specTwo)
        features = self.feature_extraction(spectrum, spectrumTwo)
        return features

    def eiip_fourier(self):
        # if len(self.read) > 0:
        spectrum = []
        spectrumTwo = []
        eiip = []
        for nucle in self.read:
            if nucle == "T" or nucle == "U":
                eiip.append(0.1335)
            elif nucle == "C":
                eiip.append(0.1340)
            elif nucle == "A":
                eiip.append(0.1260)
            else:
                eiip.append(0.0806)
        # if len(eiip) == 0:
        #     eiip = [0]
        Feiip = fft(eiip)
        for i in range(len(self.read)):
            specTotal = (abs(Feiip[i])**2)
            specTwo = (abs(Feiip[i]))
            spectrum.append(specTotal)
            spectrumTwo.append(specTwo)
        features = self.feature_extraction(spectrum, spectrumTwo)
        return features

    def complex_number_fourier(self):
        # if len(self.read) > 0:
        spectrum = []
        spectrumTwo = []
        complexNumber = []
        for nucle in self.read:
            if nucle == "T" or nucle == "U":
                complexNumber.append(1-1j)
            elif nucle == "C":
                complexNumber.append(-1+1j)
            elif nucle == "A":
                complexNumber.append(1+1j)
            else:
                complexNumber.append(-1-1j)
        # if len(complexNumber) == 0:
        #     complexNumber = [0]
        FcomplexNumber = fft(complexNumber)
        for i in range(len(self.read)):
            specTotal = (abs(FcomplexNumber[i])**2)
            specTwo = (abs(FcomplexNumber[i]))
            spectrum.append(specTotal)
            spectrumTwo.append(specTwo)
        features = self.feature_extraction(spectrum, spectrumTwo)
        return features

    def atomic_number_fourier(self):
        # if len(self.read) > 0:
        spectrum = []
        spectrumTwo = []
        atomicNumber = []
        for nucle in self.read:
            if nucle == "T" or nucle == "U":
                atomicNumber.append(66)
            elif nucle == "C":
                atomicNumber.append(58)
            elif nucle == "A":
                atomicNumber.append(70)
            else:
                atomicNumber.append(78)
        # if len(FatomicNumber) == 0:
        #     FatomicNumber = [0]
        FatomicNumber = fft(atomicNumber)
        for i in range(len(self.read)):
            specTotal = (abs(FatomicNumber[i])**2)
            specTwo = (abs(FatomicNumber[i]))
            spectrum.append(specTotal)
            spectrumTwo.append(specTwo)
        features = self.feature_extraction(spectrum, spectrumTwo)
        return features

    def Encoding(self):
        # ed1 = self.binary_fourier()
        ed2 = self.atomic_number_fourier()
        # ed3 = self.complex_number_fourier()
        # ed4 = self.eiip_fourier()
        # ed5 = self.integer_fourier()
        # ed6 = self.zcurve_fourier()
        # ed7 = self.real_fourier()
        # encoding_vector = ed1 + ed2 + ed3 + ed4 + ed5 + ed6 + ed7
        # return encoding_vector
        return ed2


#######################################################################################################
class ChaosGame():
    def __init__(self, read, max_len, kmer_freq):
        '''
        kmer_freq: Frequency of k-mer (e.g., 3, 4)
        '''
        self.read = read.upper()
        self.max_len = max_len
        # self.padding_mode = padding_mode
        self.kmer_freq = kmer_freq
        
    def chunksTwo(self, seq, win):
        seqlen = len(seq)
        for i in range(seqlen):
            j = seqlen if i+win>seqlen else i+win
            yield seq[i:j]
            if j==seqlen: break
        return

    def frequency_chaos(self, k):
        # if len(self.read) >= k:
        mapping = []
        kmer = {}
        totalWindows = (len(self.read) - k) + 1 # (L - k + 1)
        for subseq in self.chunksTwo(self.read, k):
            # print(subseq)
            if subseq in kmer:
                kmer[subseq] = kmer[subseq] + 1
            else:
                kmer[subseq] = 1
        for subseq in self.chunksTwo(self.read, k):

            mapping.append(kmer[subseq]/totalWindows)
 
        padding = ((self.max_len - k + 1) - len(mapping))

        # mapping = np.pad(mapping, (0, padding), self.padding_mode)
        return mapping + [0]*padding

    def classifical_chaos(self):
        # if len(self.read) > 0:
        Sx = []
        Sy = []
        for nucle in self.read:
            if nucle == "A":
                Sx.append(1)
                Sy.append(1)
            elif nucle == "C":
                Sx.append(-1)
                Sy.append(-1)
            elif nucle == "T" or nucle =="U":
                Sx.append(-1)
                Sy.append(1)
            else:
                Sx.append(1)
                Sy.append(-1)
        CGR_x = [] 
        CGR_y = []
        for i in range(0, len(Sx)):
            if i == 0:
                CGR_x.append(0.5 * Sx[i])
                CGR_y.append(0.5 * Sy[i])
            else:
                CGR_x.append(0.5 * Sx[i] + 0.5 * CGR_x[i - 1])
                CGR_y.append(0.5 * Sy[i] + 0.5 * CGR_y[i - 1])           
        concat = CGR_x + CGR_y
        # if len(concat) == 0:
        #     concat = [0]
        padding = (self.max_len - len(Sx)) * 2
        # mapping = np.pad(concat, (0, padding), self.padding_mode)
        return concat + [0]*padding

    def classifical_chaos_fourier(self):
        # if len(self.read) > 0:
        spectrum = []
        spectrumTwo = []
        Sx = []
        Sy = []
        for nucle in self.read:
            if nucle == "A":
                Sx.append(1)
                Sy.append(1)
            elif nucle == "C":
                Sx.append(-1)
                Sy.append(-1)
            elif nucle == "T" or nucle =="U":
                Sx.append(-1)
                Sy.append(1)
            else:
                Sx.append(1)
                Sy.append(-1)
        CGR_x = [] 
        CGR_y = []
        for i in range(0, len(Sx)):
            if i == 0:
                CGR_x.append(0.5 * Sx[i])
                CGR_y.append(0.5 * Sy[i])
            else:
                CGR_x.append(0.5 * Sx[i] + 0.5 * CGR_x[i - 1])
                CGR_y.append(0.5 * Sy[i] + 0.5 * CGR_y[i - 1])      
        Fx = fft(CGR_x)
        Fy = fft(CGR_y)
        for i in range(len(Fx)):
            specTotal = (abs(Fx[i])**2) + (abs(Fy[i])**2)
            specTwo = (abs(Fx[i])) + (abs(Fy[i]))
            spectrum.append(specTotal)
            spectrumTwo.append(specTwo)
        features = FourierTransform(self.read).feature_extraction(spectrum, spectrumTwo)
        # else:
        #     features = np.zeros(19)
        return features

    def frequency_chaos_fourier(self, k):
        # if len(self.read) > 0:
        spectrum = []
        spectrumTwo = []
        mapping = []
        kmer = {}
        totalWindows = (len(self.read) - k) + 1 # (L - k + 1)
        for subseq in self.chunksTwo(self.read, k):
            # print(subseq)
            if subseq in kmer:
                kmer[subseq] = kmer[subseq] + 1
            else:
                kmer[subseq] = 1
        for subseq in self.chunksTwo(self.read, k):
            # print(kmer[subseq])
            # print(kmer[subseq]/totalWindows)
            mapping.append(kmer[subseq]/totalWindows)
        # if len(mapping) == 0:
        #     mapping = [0]
        Fmap = fft(mapping)
        for i in range(len(mapping)):
            specTotal = (abs(Fmap[i])**2)
            specTwo = (abs(Fmap[i]))
            spectrum.append(specTotal)
            spectrumTwo.append(specTwo)
        features = FourierTransform(self.read).feature_extraction(spectrum, spectrumTwo)
        # else:
        #     features = np.zeros(19)
        return features

    def Encoding(self):
        # ed1 = self.classifical_chaos()
        ed2 = self.classifical_chaos_fourier()
        # ed3 = self.frequency_chaos(self.kmer_freq)
        ed4 = self.frequency_chaos_fourier(self.kmer_freq)
        # encoding_vector = np.concatenate((ed1, ed2, ed3, ed4), axis=None)
        # return encoding_vector
        # return ed1 + ed2 + ed3 + ed4
        return ed2 + ed4

#######################################################################################################
class Entropy():
    def __init__(self, read, kmer, q=2):
        '''
        kmer: 'Range of k-mer, E.g., 1-mer (1) or 2-mer (1, 2) ...'
        q: 'Tsallis - entropic parameter q'
        '''
        self.read = read.upper()
        self.kmer = kmer
        self.q = q

    def chunks(seq, win, step):
        seqlen = len(seq)
        for i in range(0,seqlen,step):
            j = seqlen if i+win>seqlen else i+win
            yield seq[i:j]
            if j==seqlen: break
        return         

    def chunks_two(self, seq, win):
        seqlen = len(seq)
        for i in range(seqlen):
            j = seqlen if i+win>seqlen else i+win
            yield seq[i:j]
            if j==seqlen: break
        return

    def entropy_equation(self, entropy_type):
        information_entropy = []
        for k in range(1, self.kmer+1):
            probabilities = []
            kmer = {}
            total_windows = (len(self.read) - k) + 1 # (L - k + 1)
            for subseq in self.chunks_two(self.read, k):
                if subseq in kmer:
                    # print(subseq)
                    kmer[subseq] = kmer[subseq] + 1
                else:
                    kmer[subseq] = 1
            for key, value in kmer.items():
                # print(key)
                # print(value)
                probabilities.append(value/total_windows)
            if entropy_type == "Shannon" or entropy_type == "shannon":
                entropy_equation = [(p * math.log(p, 2)) for p in probabilities]
                entropy = -(sum(entropy_equation))
                information_entropy.append(entropy)
            elif entropy_type == "Tsallis" or entropy_type == "tsallis":
                # q = 2
                entropy_equation = [(p ** self.q) for p in probabilities]
                entropy =  (1/(self.q - 1)) * (1 - sum(entropy_equation))
                information_entropy.append(entropy)
        return information_entropy  

    def Encoding(self):
        ed1 = self.entropy_equation(entropy_type = "tsallis")
        ed2 = self.entropy_equation(entropy_type = "shannon")
        return ed1 + ed2

###############################################################################
class FickettScore():
    def __init__(self, read, read_type = 'DNA'):
        '''
        kmer_freq: Frequency of k-mer (e.g., 3, 4)
        '''
        self.read = read.upper()
        self.read_type = read_type

    def look_up_position_prob(self, value, base, position_para, position_prob, position_weight):
        """look up positional probability by base and value"""

        if float(value) < 0:
            return None
        for idx, val in enumerate(position_para):
            if float(value) >= val:
                return float(position_prob[base][idx]) * float(position_weight[base])

    def look_up_content_prob(self, value, base, content_para, content_prob, content_weight):

        """look up content probability by base and value"""

        if float(value) < 0:
            return None
        for idx, val in enumerate(content_para):
            if float(value) >= val:
                return float(content_prob[base][idx]) * float(content_weight[base])

    def fickett_value_orf(self, seq):

        """calculate Fickett value. Input is DNA sequence"""

        position_prob = {
            'A': [0.94, 0.68, 0.84, 0.93, 0.58, 0.68, 0.45, 0.34, 0.20, 0.22],
            'C': [0.80, 0.70, 0.70, 0.81, 0.66, 0.48, 0.51, 0.33, 0.30, 0.23],
            'G': [0.90, 0.88, 0.74, 0.64, 0.53, 0.48, 0.27, 0.16, 0.08, 0.08],
            'T': [0.97, 0.97, 0.91, 0.68, 0.69, 0.44, 0.54, 0.20, 0.09, 0.09]}

        position_weight = {'A': 0.26, 'C': 0.18, 'G': 0.31, 'T': 0.33}
        position_para = [1.9, 1.8, 1.7, 1.6, 1.5, 1.4, 1.3, 1.2, 1.1, 0.0]

        content_prob = {
            'A': [0.28, 0.49, 0.44, 0.55, 0.62, 0.49, 0.67, 0.65, 0.81, 0.21],
            'C': [0.82, 0.64, 0.51, 0.64, 0.59, 0.59, 0.43, 0.44, 0.39, 0.31],
            'G': [0.40, 0.54, 0.47, 0.64, 0.64, 0.73, 0.41, 0.41, 0.33, 0.29],
            'T': [0.28, 0.24, 0.39, 0.40, 0.55, 0.75, 0.56, 0.69, 0.51, 0.58]}

        content_weight = {'A': 0.11, 'C': 0.12, 'G': 0.15, 'T': 0.14}
        content_para = [0.33, 0.31, 0.29, 0.27, 0.25, 0.23, 0.21, 0.17, 0]

        if len(seq) < 2:
            return 0
        fickett_score = 0
        seq = seq.upper()
        total_base = len(seq)
        A_content = float(seq.count('A')) / total_base
        C_content = float(seq.count('C')) / total_base
        G_content = float(seq.count('G')) / total_base
        if self.read_type == 'DNA':
            T_content = float(seq.count('T')) / total_base
        elif self.read_type == 'RNA':
            T_content = float(seq.count('U')) / total_base

        phase_0 = [seq[i] for i in range(0, len(seq)) if i % 3 == 0]
        phase_1 = [seq[i] for i in range(0, len(seq)) if i % 3 == 1]
        phase_2 = [seq[i] for i in range(0, len(seq)) if i % 3 == 2]
        
        A_position = max(phase_0.count('A'), phase_1.count('A'), phase_2.count('A')) / (min(phase_0.count('A'), phase_1.count('A'), phase_2.count('A')) + 1.0)

        C_position = max(phase_0.count('C'), phase_1.count('C'), phase_2.count('C')) / (min(phase_0.count('C'), phase_1.count('C'), phase_2.count('C')) + 1.0)

        G_position = max(phase_0.count('G'), phase_1.count('G'), phase_2.count('G')) / (min(phase_0.count('G'), phase_1.count('G'), phase_2.count('G')) + 1.0)

        if self.read_type == 'DNA':
            T_position = max(phase_0.count('T'), phase_1.count('T'), phase_2.count('T')) / (min(phase_0.count('T'), phase_1.count('T'), phase_2.count('T')) + 1.0)
        elif self.read_type == 'RNA':
            T_position = max(phase_0.count('U'), phase_1.count('U'), phase_2.count('U')) / (min(phase_0.count('U'), phase_1.count('U'), phase_2.count('U')) + 1.0)

        fickett_score += self.look_up_content_prob(A_content, 'A', content_para, content_prob, content_weight)
        fickett_score += self.look_up_content_prob(C_content, 'C', content_para, content_prob, content_weight)
        fickett_score += self.look_up_content_prob(G_content, 'G', content_para, content_prob, content_weight)
        fickett_score += self.look_up_content_prob(T_content, 'T', content_para, content_prob, content_weight)
        
        fickett_score += self.look_up_position_prob(A_position, 'A', position_para, position_prob, position_weight)
        fickett_score += self.look_up_position_prob(C_position, 'C', position_para, position_prob, position_weight)
        fickett_score += self.look_up_position_prob(G_position, 'G', position_para, position_prob, position_weight)
        fickett_score += self.look_up_position_prob(T_position, 'T', position_para, position_prob, position_weight)  
        return fickett_score

    def fickett_value_full_sequence(self, seq):

        """calculate Fickett from full sequence - CPC2"""

        position_para = [1.9, 1.8, 1.7, 1.6, 1.5, 1.4, 1.3, 1.2, 1.1, 0.0]
        content_para = [0.33, 0.31, 0.29, 0.27, 0.25, 0.23, 0.21, 0.19, 0.17, 0]

        position_prob = {
            'A': [0.51, 0.55, 0.57, 0.52, 0.48, 0.58, 0.57, 0.54, 0.50, 0.36],
            'C': [0.29, 0.44, 0.55, 0.49, 0.52, 0.60, 0.60, 0.56, 0.51, 0.38],
            'G': [0.62, 0.67, 0.74, 0.65, 0.61, 0.62, 0.52, 0.41, 0.31, 0.17],
            'T': [0.51, 0.60, 0.69, 0.64, 0.62, 0.67, 0.58, 0.48, 0.39, 0.24]}
        
        position_weight = {'A': 0.062, 'C': 0.093, 'G': 0.205, 'T': 0.154}
        content_weight = {'A': 0.084, 'C': 0.076, 'G': 0.081, 'T': 0.055}

        content_prob = {
            'A': [0.40, 0.55, 0.58, 0.58, 0.52, 0.48, 0.45, 0.45, 0.38, 0.19],
            'C': [0.50, 0.63, 0.59, 0.50, 0.46, 0.45, 0.47, 0.56, 0.59, 0.33],
            'G': [0.21, 0.40, 0.47, 0.50, 0.52, 0.56, 0.57, 0.52, 0.44, 0.23],
            'T': [0.30, 0.49, 0.56, 0.53, 0.48, 0.48, 0.52, 0.57, 0.60, 0.51]}

        if len(seq) < 2:
            return 0

        fickett_score = 0
        seq = seq.upper()
        total_base = len(seq)

        phase_0 = seq[::3]
        phase_1 = seq[1::3]
        phase_2 = seq[2::3]

        phase_0_A = phase_0.count('A')
        phase_1_A = phase_1.count('A')
        phase_2_A = phase_2.count('A')
        phase_0_C = phase_0.count('C')
        phase_1_C = phase_1.count('C')
        phase_2_C = phase_2.count('C')
        phase_0_G = phase_0.count('G')
        phase_1_G = phase_1.count('G')
        phase_2_G = phase_2.count('G')
        if self.read_type == 'DNA':
            phase_0_T = phase_0.count('T')
            phase_1_T = phase_1.count('T')
            phase_2_T = phase_2.count('T')
        elif self.read_type == 'RNA':
            phase_0_T = phase_0.count('U')
            phase_1_T = phase_1.count('U')
            phase_2_T = phase_2.count('U')

        A_content = float(phase_0_A + phase_1_A + phase_2_A) / total_base
        C_content = float(phase_0_C + phase_1_C + phase_2_C) / total_base
        G_content = float(phase_0_G + phase_1_G + phase_2_G) / total_base
        T_content = float(phase_0_T + phase_1_T + phase_2_T) / total_base
        A_position = max([phase_0_A, phase_1_A, phase_2_A]) / (min([phase_0_A, phase_1_A, phase_2_A]) + 1.0)
        C_position = max([phase_0_C, phase_1_C, phase_2_C]) / (min([phase_0_C, phase_1_C, phase_2_C]) + 1.0)
        G_position = max([phase_0_G, phase_1_G, phase_2_G]) / (min([phase_0_G, phase_1_G, phase_2_G]) + 1.0)
        T_position = max([phase_0_T, phase_1_T, phase_2_T]) / (min([phase_0_T, phase_1_T, phase_2_T]) + 1.0)

        fickett_score += self.look_up_content_prob(A_content, 'A', content_para, content_prob, content_weight)
        fickett_score += self.look_up_content_prob(C_content, 'C', content_para, content_prob, content_weight)
        fickett_score += self.look_up_content_prob(G_content, 'G', content_para, content_prob, content_weight)
        fickett_score += self.look_up_content_prob(T_content, 'T', content_para, content_prob, content_weight)

        fickett_score += self.look_up_position_prob(A_position, 'A', position_para, position_prob, position_weight)
        fickett_score += self.look_up_position_prob(C_position, 'C', position_para, position_prob, position_weight)
        fickett_score += self.look_up_position_prob(G_position, 'G', position_para, position_prob, position_weight)
        fickett_score += self.look_up_position_prob(T_position, 'T', position_para, position_prob, position_weight)

        return fickett_score

    def calculate_sequences(self):
        measure_orf = self.fickett_value_orf(self.read)
        measure_full = self.fickett_value_full_sequence(self.read)
        return [measure_orf, measure_full]

class EncodeScheme():
    def __init__(self, max_len, entropy_kmer = 3, entropy_q=2, kmer_freq = 3, read_type='DNA'):
        super(EncodeScheme, self).__init__()
        self.max_len = max_len
        self.entropy_kmer = entropy_kmer
        self.entropy_q = entropy_q
        self.kmer_freq = kmer_freq        
        self.read_type = read_type

    def atomic_number(self, read):
        mapping = []
        for nucle in read:
            if nucle == "T" or nucle == "U":
                mapping.append(66)
                # mapping.append(10)
            elif nucle == "C":
                mapping.append(58)
                # mapping.append(20)
            elif nucle == "A":
                mapping.append(70)
                # mapping.append(30)
            elif nucle == "G":
                mapping.append(78)
                # mapping.append(40)
            else:
                mapping.append(-1)
        padding = self.max_len + 1 - len(read)
        mapping = mapping + [0]*padding
        return mapping

    def binary(self, read):
        mapping = []
        for nucle in read:
            if nucle == "T" or nucle == "U":
                mapping.extend([1, 0, 0, 0])
                # mapping.append(10)
            elif nucle == "C":
                mapping.extend([0, 1, 0, 0])
                # mapping.append(20)
            elif nucle == "A":
                mapping.extend([0, 0, 1, 0])
                # mapping.append(30)
            elif nucle == "G":
                mapping.extend([0, 0, 0, 1])
                # mapping.append(40)
            else:
                mapping.extend([-1, -1, -1, -1])
        padding = self.max_len + 1 - len(read)
        mapping = mapping + [0, 0, 0, 0]*padding
        return mapping

    def descriptors(self, method, read):
        if method == "FourierTransform":
            features = FourierTransform(read).Encoding()
        elif method == "ChaosGame":
            features = ChaosGame(read, self.max_len, self.kmer_freq).Encoding()
        elif method == "Entropy":
            features = Entropy(read, self.entropy_kmer, self.entropy_q).Encoding()
        elif method == "FickettScore": # dim = 2
            features = FickettScore(read, self.read_type).calculate_sequences()
        elif method == "binary":
            features = self.binary(read)
        elif method == "atomic_number":
            features = self.atomic_number(read)
        return features


'''
Author: Pengyao PING
Date: 2022-06-03 01:36:38
LastEditors: Pengyao PING
LastEditTime: 2022-08-04 00:26:03
Email: Pengyao.Ping@student.uts.edu.au
Description: 
'''
import unittest
from noise2read.reads2vectors import Reads2Vectors

class TestReads2Vectors(unittest.TestCase):
    def __init__(self):
        data_dir = '/data/pping/Repo/srec/data/test'
        lab_csv = '/data/pping/Repo/srec/data/test/test.csv'
        unlab_csv = '/data/pping/Repo/srec/data/test/test.csv'
        k_lst = '1, 2, 3'
        self.RV = Reads2Vectors(data_dir, lab_csv, unlab_csv, k_lst)
    
    def test_doubling_kmer_contribution(self):
        read = 'ACGTACTGAACT'
        k = 3
        kmers1 = ['ACG', 'ACG']
        res = self.RV.doubling_kmer_contribution(read, k, kmers1)
        self.assertEqual(res, 0)

        kmers2 = ['ACT', 'ACT']
        res2 = self.RV.doubling_kmer_contribution(read, k, kmers2)
        self.assertEqual(res2, 0)


if __name__ == '__main__':
    unittest.main()


# -*- coding: utf-8 -*-
# @Author: Pengyao Ping
# @Date:   2023-02-16 11:01:06
# @Last Modified by:   Pengyao Ping
# @Last Modified time: 2023-02-16 11:10:34
import os
from noise2read.utils import *

class SequenceCoverage():
    """
    A class to calculate sequence coverage using read datasets and genome.
    """
    def __init__(self, logger, config):
        """
        initialize the SequenceCoverage class

        Args:
            logger (class): customized logging
            config (class): parameters setting using configparser
        """
        self.logger = logger
        self.config = config
        if config.library_layout == "single" or "SR":
            self.logger.info(f'Read Set: {config.read}')
            bases = config.read.split('/')[-1]
            self.prefix = bases.split('.' + parse_file_type(config.read))[0]
        elif config.library_layout == "paired" or "PE":
            self.logger.info(f'Read Set 1: {config.read1}')
            self.logger.info(f'Read Set 2: {config.read2}')
            bases = config.read1.split('/')[-1]
            self.prefix = bases.split('.' + parse_file_type(config.read1))[0]
        if os.path.exists(self.config.result_dir):
            self.logger.info("Directory '% s' already exists, noise2read will use it" % self.config.result_dir)
        else:
            os.makedirs(self.config.result_dir)
            self.logger.info("Directory '% s' created" % self.config.result_dir)

    def build_index(self, reference):
        self.logger.info("Building genome index using bowtie2")
        ref_index_dir = os.path.join(self.config.result_dir, '/reference_index')
        if os.path.exists(ref_index_dir):
            self.logger.info("Directory '% s' already exists, noise2read will use it" % ref_index_dir)
        else:
            os.makedirs(ref_index_dir)
            self.logger.info("Directory '% s' created" % ref_index_dir)
        os.system("bowtie2-build --large-index %s %s" % (reference, ref_index_dir))
        return ref_index_dir

    def alignment(self, ref_index_dir):
        self.logger.info("Alignment using bowtie2")
        align_result = os.path.join(self.config.result_dir, self.prefix + '_alginment.sam')
        if self.config.library_layout == "single" or "SR":
            os.system("bowtie2 %s -p %s -x %s -0 %s > %s" % (self.config.Alignment, self.config.num_workers, ref_index_dir, self.config.read, align_result))
        elif self.config.library_layout == "paired" or "PE":
            os.system("bowtie2 %s -p %s -x %s -1 %s -2 %s > %s" % (self.config.Alignment, self.config.num_workers, ref_index_dir, self.config.read1, self.config.read2, align_result))
        return align_result

    


'''
Author: Pengyao PING
Date: 2022-05-27 14:08:11
LastEditors: Pengyao PING
LastEditTime: 2022-08-04 00:26:06
Email: Pengyao.Ping@student.uts.edu.au
Description: 
'''
import unittest
from unittest import result
from noise2read.data_generation import DataGneration

class TestDataGeneration(unittest.TestCase):
    def __init__(self, methodName: str = ...) -> None:
        super().__init__(methodName)
        input_filename = '/data/pping/Repo/srec/data/test/test.fastq'
        file_type = 'fastq'
        output_dir = '/data/pping/Repo/srec/data/test'
        self.DG = DataGneration(input_filename, file_type, output_dir, max_error_freq = 2)
    def test_sub_base(self):
        self.assertEqual('TCG', self.DG.sub_base('A'))  
        self.assertEqual('ACG', self.DG.sub_base('T'))
        self.assertEqual('ATG', self.DG.sub_base('C'))
        self.assertEqual('ACT', self.DG.sub_base('G'))   
        self.assertEqual('N', self.DG.sub_base('N'))    

    def test_seq2deletion(self):
        read = 'ACGTGA'
        del_list = self.DG.seq2deletion(read)
        self.assertEqual(del_list, set(["CGTGA", 'AGTGA', 'ACTGA', 'ACGGA', 'ACGTA','ACGTG']))

    def test_seq2insertion(self):
        read = 'ACGT'
        read_set = self.DG.seq2insertion(read)
        exp_result = set(['AACGT', 'GACGT', 'CACGT', 'TACGT', 'AACGT', 'AGCGT','ACCGT', 'ATCGT', 'ACAGT', 'ACGGT', 'ACCGT', 'ACTGT', 'ACGAT', 'ACGGT', 'ACGCT', 'ACGTT', 'ACGTA', 'ACGTG', 'ACGTC', 'ACGTT'])
        self.assertEqual(read_set, exp_result)     

    def test_seq2substitution(self):
        read1 = 'ACGT'
        del_list1 = self.DG.seq2substitution(read1)
        exp_result1 = set(['CCGT', 'GCGT', 'TCGT', 'AAGT', 'AGGT', 'ATGT', 'ACAT', 'ACCT', 'ACTT', 'ACGA', 'ACGG', 'ACGC'])
        self.assertEqual(del_list1, exp_result1)     
        read2 = 'ACGN'
        del_list2 = self.DG.seq2substitution(read2)
        exp_result2 = set(['CCGN', 'GCGN', 'TCGN', 'AAGN', 'AGGN', 'ATGN', 'ACAN', 'ACCN', 'ACTN'])
        self.assertEqual(del_list2, exp_result2)    

    def test_ErrorTypeClassification(self):
        line1 = ['ACGTA', 5, 'ACGTAA', 1]
        real_result1 = self.DG.ErrorTypeClassification(line1)
        exp_result1 = ['ACGTA', 5, 'N-A', 5, 'ACGTAA', 1]
        self.assertEqual(real_result1, exp_result1) 

        line1 = ['ACGTAG', 5, 'ACGTAA', 1]
        real_result1 = self.DG.ErrorTypeClassification(line1)
        exp_result1 = ['ACGTAG', 5, 'G-A', 5, 'ACGTAA', 1]
        self.assertEqual(real_result1, exp_result1) 

        line1 = ['TACGTA', 5, 'ACGTA', 1]
        real_result1 = self.DG.ErrorTypeClassification(line1)
        exp_result1 = ['TACGTA', 5, 'T-N', 0, 'ACGTA', 1]
        self.assertEqual(real_result1, exp_result1) 

        # line1 = ['TACG', 5, 'ACGTA', 1]
        # # real_result1 = self.DG.ErrorTypeClassification(line1)
        # # raise IOError("The editdistance of two reads in the input list must equal to one!")
        # self.assertRaises(IOError, self.DG.ErrorTypeClassification(line1)) 

    def test_EditDis1Seqs(self):
        read1 = 'ACGTA'
        del_list1 = self.DG.EditDis1Seqs(read1, 'A-G')
        exp_result1 = set(['GCGTA', 'ACGTG'])
        self.assertEqual(del_list1, exp_result1)     
        read2 = 'ACGT'
        del_list2 = self.DG.EditDis1Seqs(read2, 'N-A')
        exp_result2 = set(['AACGT', 'ACAGT', 'ACGAT', 'ACGTA'])
        self.assertEqual(del_list2, exp_result2) 

        read2 = 'ACGTT'
        del_list2 = self.DG.EditDis1Seqs(read2, 'T-N')
        exp_result2 = set(['ACGT', 'ACGT'])
        self.assertEqual(del_list2, exp_result2) 

if __name__ == '__main__':
    unittest.main()


# -*- coding: utf-8 -*-
# @Author: Pengyao Ping
# @Date:   2023-02-16 11:01:06
# @Last Modified by:   Pengyao Ping
# @Last Modified time: 2023-05-16 18:27:41

import optuna
from optuna.samplers import TPESampler
from sklearn.model_selection import train_test_split
import os
import xgboost as xgb
import numpy as np
# from imblearn.combine import SMOTEENN
from imblearn.over_sampling import SMOTE 
import collections
# from imblearn.under_sampling import RandomUnderSampler
from matplotlib import pyplot
from sklearn.metrics import f1_score
import optuna
import numpy as np
# from matplotlib import pyplot as plt
# from sklearn.metrics import classification_report, confusion_matrix
from noise2read.utils import MemoryMonitor

class MLClassifier:
    """
    A classifier based on xgboost and optuna to predict whether a high-/low-frequency read is an error read mutated from a read when sequencing.
    """
    def __init__(self, logger, config, study_name, data, label, ambi_data):
        """
        initialize the MLClassifier class

        Args:
            logger (class): customized logging
            config (class): parameters setting using configparser
            study_name (str): study name of optuna
            data (arrays): numpy arrays with same length / shape[0]
            label (arrays): numpy arrays, sample labels (n_samples, n_classes(0/1))
            ambi_data (arrays): numpy arrays with same length / shape[0]
        """
        self.logger = logger
        self.config = config
        self.study_name = study_name
        self.data = data
        self.label = label
        self.ambi_data = ambi_data
        self.MM = MemoryMonitor(self.logger)

        # Shuffle the data in a deterministic way
        np.random.seed(self.config.random_state)
        indices = np.random.permutation(len(self.data))
        shuffled_data = self.data[indices]
        shuffled_labels = self.label[indices]
        self.x_train, self.x_test, self.y_train, self.y_test = train_test_split(shuffled_data, shuffled_labels, test_size=self.config.test_size, shuffle=False, random_state=self.config.random_state)

        # self.x_train, self.x_test, self.y_train, self.y_test = train_test_split(self.data, self.label, test_size=self.config.test_size, shuffle=True, random_state=self.config.random_state)

        # self.x_test, self.x_val, self.y_test, self.y_val, self.x_test_weight, self.x_val_weight = train_test_split(test_val_x, test_val_y, test_val_weight, test_size=0.50, random_state=self.config.random_state, shuffle=True)
        # self.x_test_weight, self.x_val_weight = train_test_split(test_val_weight, test_size=0.50, random_state=self.config.random_state, shuffle=False)
        # self.logger.debug(type(self.x_val_weight))
        # self.logger.debug(f'{self.x_val}, {self.y_val.shape}, {len(self.x_val_weight)}')
        self.logger.info(f'The number of negative and positive samples: {sorted(collections.Counter(self.y_train).items())}')

        sm = SMOTE(random_state=self.config.random_state)
        self.X_resampled, self.y_resampled = sm.fit_resample(self.x_train, self.y_train)
        self.logger.info(f'After over-sampling: {sorted(collections.Counter(self.y_resampled).items())}')
        del sm
        self.MM.measure()
        # rus = RandomUnderSampler(random_state=42)
        # self.X_resampled, self.y_resampled = rus.fit_resample(self.x_train, self.y_train)
        # self.logger.info(f'After under-sampling {sorted(collections.Counter(self.y_resampled).items())}')

        # smote_enn = SMOTEENN(random_state=0)
        # self.X_resampled, self.y_resampled = smote_enn.fit_resample(self.x_train, self.y_train)
        # self.logger.info(f'Combination of over- and under-sampling {sorted(collections.Counter(self.y_resampled).items())}')
 
    def objective(self, trial):
        """
        Define an objective function to be minimized in the framewokr of Optuna

        Args:
            trial (class): The trial module of Optuna contains Trial related classes and functions. A Trial instance represents a process of evaluating an objective function. More details please see Optuna API reference: https://optuna.readthedocs.io/en/stable/reference/trial.html
            
        Raises:
            optuna.TrialPruned: Exception for pruned trials of Optuna, more details see https://optuna.readthedocs.io/en/stable/reference/generated/optuna.TrialPruned.html#optuna.TrialPruned

        Returns:
            float: test data accuracy
        """
        # X, y = shuffle(self.data, self.label)
        param = {
            "verbosity": 0,
            # 'gpu_id': 0,
            # "max_delta_step": 1,
            "tree_method": self.config.tree_method,
            "objective": "binary:logistic",
            'n_estimators': self.config.n_estimators,
            'early_stopping_rounds': 100,
            # "eval_metric": ['auc', 'error'],
            # "eval_metric": ['logloss'],
            "eval_metric": ['auc', 'logloss'],
            # "booster": trial.suggest_categorical("booster", ["gbtree", "dart"]),
            "booster": trial.suggest_categorical("booster", ["gbtree", "gblinear", "dart"]),
            "lambda": trial.suggest_float("lambda", 1e-8, 1.0, log=False),
            "alpha": trial.suggest_float("alpha", 1e-8, 1.0, log=False),
            "learning_rate": trial.suggest_float("learning_rate", self.config.learning_rate_min, self.config.learning_rate_max, log=False),
            "max_depth": trial.suggest_int("max_depth", self.config.max_depth_min, self.config.max_depth_max, step = self.config.max_depth_step, log = False),
            "subsample": trial.suggest_float('subsample', self.config.subsample_min, self.config.subsample_max),
            "colsample_bytree": trial.suggest_float('colsample_bytree', self.config.colsample_bytree_min, self.config.colsample_bytree_max),
            # the following setting only for gbtree and dart
            # "eta": trial.suggest_float("eta", 1e-8, 1.0, log=False),
            # "gamma": trial.suggest_float("gamma", 1e-8, 1.0, log=False),
            # "grow_policy": trial.suggest_categorical("grow_policy", ["depthwise", "lossguide"]),
            # "num_boost_round":  trial.suggest_int("num_boost_round", self.config.num_boost_round_min, self.config.num_boost_round_max, step = self.config.num_boost_round_step, log=False)
            # "num_boost_round": 50
        }

        if param["booster"] == "gbtree" or param["booster"] == "dart":
            param["max_depth"] = trial.suggest_int("max_depth", self.config.max_depth_min, self.config.max_depth_max, step = self.config.max_depth_step, log = False)
            param["eta"] = trial.suggest_float("eta", 1e-8, 1.0, log=False)
            param["gamma"] = trial.suggest_float("gamma", 1e-8, 1.0, log=False)
            param["grow_policy"] = trial.suggest_categorical("grow_policy", ["depthwise", "lossguide"])
        if param["booster"] == "dart":
            param["sample_type"] = trial.suggest_categorical("sample_type", ["uniform", "weighted"])
            param["normalize_type"] = trial.suggest_categorical("normalize_type", ["tree", "forest"])
            param["rate_drop"] = trial.suggest_float("rate_drop", 1e-8, 1.0, log=False)
            param["skip_drop"] = trial.suggest_float("skip_drop", 1e-8, 1.0, log=False)
        
        pruning_callback = optuna.integration.XGBoostPruningCallback(trial, "validation_0-auc")
        # pruning_callback = optuna.integration.XGBoostPruningCallback(trial, "validation_1-logloss")
        # pruning_callback = optuna.integration.XGBoostPruningCallback(trial, "validation_1-aucpr")

        # y_count = sorted(collections.Counter(self.y_train).items())
        # _, neg_num = y_count[0]
        # _, pos_num = y_count[1]
        # xgbc = xgb.XGBClassifier(**param, pruning_callback=[pruning_callback], probability=True, scale_pos_weight= neg_num/pos_num)
        # xgbc.fit(self.x_train, self.y_train, eval_set=[(self.x_train, self.y_train),(self.x_test, self.y_test)],  verbose=True)
        # train_accuracy = xgbc.score(self.x_train, self.y_train)

        xgbc = xgb.XGBClassifier(**param, pruning_callback=[pruning_callback], probability=True, seed=self.config.xgboost_seed)
        xgbc.fit(self.X_resampled, self.y_resampled, eval_set=[(self.X_resampled, self.y_resampled), (self.x_test, self.y_test)],  verbose=self.config.verbose_eval)
        
        # xgbc.fit(self.x_train, self.y_train, eval_set=[(self.x_train, self.y_train),(self.x_test, self.y_test)],  verbose=True)
        # xgbc.fit(self.x_train, self.y_train, eval_set=[(self.x_val, self.y_val)], verbose=False)
        # xgbc.fit(self.x_train, self.y_train, sample_weight = self.x_train_weight, eval_set=[(self.x_val, self.y_val)], sample_weight_eval_set = [self.x_val_weight],  verbose=False)

        # accuracy = xgbc.score(self.x_test, self.y_test, sample_weight=self.x_test_weight)
        # train_accuracy = xgbc.score(self.X_resampled, self.y_resampled)
        train_accuracy = xgbc.score(self.x_train, self.y_train)
        test_accuracy = xgbc.score(self.x_test, self.y_test)
        # preds = xgbc.predict(self.x_test)
        # pred_labels = np.rint(preds)
        # accuracy = accuracy_score(self.y_test, pred_labels)

        # train_f1 = f1_score(self.y_resampled, xgbc.predict(self.X_resampled))
        train_f1 = f1_score(self.y_train, xgbc.predict(self.x_train))
        test_f1 = f1_score(self.y_test, xgbc.predict(self.x_test))
        
        trial.report(test_accuracy, trial.number)
        if trial.should_prune():
            raise optuna.TrialPruned()
        
        self.logger.info( " Trial " + str(trial.number))
        self.logger.info("      Train Accuracy: {}, Test Accuracy: {}".format(train_accuracy, test_accuracy))
        self.logger.info("      Train F1: {}, Test F1: {}".format(train_f1, test_f1))
        trial.set_user_attr(key="best_model", value=xgbc) # save model

        # results = xgbc.evals_result()
        # test_logloss = results['validation_1']['logloss']
        # test_auc = results['validation_1']['auc']
        # return test_logloss
        return test_accuracy
        # return test_f1
        
    def tunning(self, n_trials = 150):
        """
        function to create a Optuna study for model training and prediction.

        Args:
            n_trials (int, optional): Number of Optuan trials. Defaults to 150.

        Returns:
            array: numpy ndarray of shape (n_samples, n_classes)
            Estimated probabilities.
        """
        # study = optuna.create_study(study_name = self.study_name,
        #     pruner=optuna.pruners.MedianPruner(n_warmup_steps=5), direction="maximize" 
        # ) #, interval_steps=10 n_startup_trials=5, 
        self.logger.info("-------------------------------------------------------------")
        sampler = TPESampler(seed=self.config.optuna_seed)  # Make the sampler behave in a deterministic way.
        study = optuna.create_study(study_name = self.study_name, direction="maximize", sampler=sampler)
        study.optimize(self.objective, n_trials, show_progress_bar=False, gc_after_trial=True)
        # print(study.best_trial)
        
        self.logger.info(f'Study Name: {self.study_name}')
        self.logger.info('Number of finished trials: {}'.format(len(study.trials)))
        self.logger.info('Best trial:')
        best_trial = study.best_trial
        self.logger.info('  Test Accuracy: {}'.format(best_trial.value))
        self.logger.info('  Params: ')
        for key, value in best_trial.params.items():
            self.logger.info('    {}: {}'.format(key, value))
        self.MM.measure()
        # best_trial_copy = copy.deepcopy(best_trial)

        # re-evaluate
        # self.objective(best_trial)
        # # # the user attribute is overwritten by re-evaluation
        # assert best_trial.user_attrs != best_trial_copy.user_attrs
        # if best_trial.value < self.config.best_accuracy:
        #     self.logger.warning(f"The stduy {self.study_name} output lower best trial accuracy of {best_trial.value} than the pre-set self.config.best_accuracy, which may result in bad error correction performance.")
        # fig1 = optuna.visualization.plot_optimization_history(study)
        # fig1.write_image(os.path.join(self.config.result_dir, "optimization_history.png"))

        # fig2 = optuna.visualization.plot_intermediate_values(study)
        # fig2.write_image(os.path.join(self.config.result_dir, "intermediate_values.png"))

        # fig3 = optuna.visualization.plot_parallel_coordinate(study)
        # fig3.write_image(os.path.join(self.config.result_dir, "parallel_coordinate.png"))

        # fig4 = optuna.visualization.plot_contour(study)
        # fig4.write_image(os.path.join(self.config.result_dir, "contour.png"))
        best_model = best_trial.user_attrs["best_model"]
        #prediction
        results = best_model.evals_result()
        
        # plot learning curves
        fig, ax = pyplot.subplots()
        
        ax.set_xlabel('estimators')
        ax.set_ylabel('logloss')
        
        ax.plot(results['validation_0']['logloss'], label='train')
        ax.plot(results['validation_1']['logloss'], label='test')
        # show the legend
        ax.legend()
        # show the plot
        fig.savefig(os.path.join(self.config.result_dir, self.study_name + '_train-test-logloss.png'))
        predictions = best_model.predict_proba(self.ambi_data)[:, 1]
        pyplot.clsoe()
        self.logger.info("-------------------------------------------------------------")
        del study, sampler, results, best_model
        self.MM.measure()
        self.MM.stop()
        return predictions


# -*- coding: utf-8 -*-
# @Author: Pengyao Ping
# @Date:   2023-01-16 15:52:44
# @Last Modified by:   Pengyao Ping
# @Last Modified time: 2023-09-05 23:40:46

import editdistance
import networkx as nx
import os
import csv
from tqdm import tqdm
from noise2read.utils import *
from mpire import WorkerPool
import matplotlib.pyplot as plt
from networkx.drawing.nx_agraph import graphviz_layout
from collections import Counter
import sys
import pandas as pd
from noise2read.utils import MemoryMonitor

class DataGneration():
    """
    A class to generate genuine and ambiguous errors from 1nt/2nt-edit-distance-based graphs construted from short reads dataset
    """
    def __init__(self, logger, config):
        """
        initialize the DataGneration class

        Args:
            logger (class): customized logging
            config (class): parameters below setting using configparser
                parameters:
                config.num_workers (int): the number of subprocesses to use.
                config.input_file (str): The filename including path to be corrected.
                config.result_dir (str): Full pathname where the results will be saved.
                config.high_freq_thre (int, optional): The threshold of the high frequency. Defaults to 4.
                config.max_error_freq (int, optional): The highest frequency of error sequences. Defaults to 3.
                config.save_graph (bool, optional): If true, noise2read will save the construted graph to file. Defaults to False. 
                config.graph_visualization (bool, optional): If true, noise2read will draw the connected subgraphs of the construted  graph. Defaults to False.
                config.drawing_graph_num (int, optional): The number of subgraphs will be drawing. Defaults to 50.
                config.high_ambiguous (bool, optional): If true, noise2read will predict high ambiguous errors. Defaults to False.
                config.min_iters (int, optional): Minimum progress display update interval in iterations for the module tqdm. Defaults to 100.
                config.verbose (bool, optional): If true, noise2read will save the genuine, ambiguous errors and negative reads to csv.
        """
        self.logger = logger
        self.config = config
        if os.path.exists(self.config.result_dir):
            self.logger.warning("Directory '%s' already exists, noise2read will use it." % self.config.result_dir)
        else:
            os.makedirs(self.config.result_dir)
            self.logger.info("Directory '%s' created" % self.config.result_dir)
        
        self.file_type = parse_file_type(self.config.input_file)
        if ".gz" in self.file_type:
            self.out_file_tye = self.file_type.split(".gz")[0] 
        else:
            self.out_file_tye = self.file_type
        # Create an instance of the MemoryMonitor
        self.MM = MemoryMonitor(self.logger)

    def graph_summary(self, graph):
        """
        print the graph info
        """
        edge_nums = graph.number_of_edges()
        node_nums = graph.number_of_nodes()
        self.logger.info("Graph Summary: Edge Number: {}, Nodes Number: {}".format(edge_nums, node_nums))
        return
        
    def err_type_classification(self, line):
        """
        calculate the error type, position, and 2- or 3-mer consisting of two bases before and after the error base.

        Args:
            line (list): [read1, read1_frequency, read1_degree, read2, read2_frequency, read2_degree]

        Raises:
            ValueError: The function raise an error when the edit distance of two reads in the input list not equal to one.

        Returns:
            list: [read1, read1_frequency, read1_degree, errorType, errorPosition, f_kmer, s_kmer, read2, read2_frequency, read2_degree]
        """
        read1 = line[0]
        read1_fre = line[1]
        read2 = line[3]   
        read2_fre = line[4] 

        f_len = len(read1)
        s_len = len(read2)  
        # position = -1  
        dis = editdistance.eval(read1, read2)
        # print(dis)
        if dis == 1:              
            if f_len == s_len:
                position = -1
                for index in range(f_len):
                    if read1[index] == read2[index]:
                        continue
                    else:
                        position = index
                        break
                first = read1[position]
                second = read2[position]
                if position == 0:
                    f_kmer = read1[0:2]
                    s_kmer = read2[0:2]
                elif position == f_len:
                    f_kmer = read1[-2:] 
                    s_kmer = read2[-2:]
                else:
                    f_kmer = read1[position-1 : position+2]
                    s_kmer = read2[position-1 : position+2]  

            elif f_len < s_len:
                position = -1
                num = 0
                for index in range(f_len):
                    if read1[index] == read2[index]:
                        num = num + 1
                    else:
                        position = index
                        break
                if num == f_len:
                    position = f_len
                first = 'X'
                second = read2[position]  
                if position == 0:
                    f_kmer = 'X' + read1[0]
                    s_kmer = read2[0:2]
                elif position == f_len:
                    f_kmer = read1[-1] + 'X'
                    s_kmer = read2[-2:]
                else:
                    f_kmer = read1[position-1] + 'X' + read1[position]
                    s_kmer = read2[position-1 : position+2]                    
            elif f_len > s_len:
                position = -1
                num = 0
                for index in range(s_len):
                    if read1[index] == read2[index]:
                        num = num + 1
                    else:
                        position = index
                        break
                if num == s_len:
                    position = s_len
                first = read1[position]
                second = 'X'
                if position == 0:
                    f_kmer = read1[0:2]
                    s_kmer = 'X' + read2[0]
                elif position == s_len:
                    f_kmer = read1[-2:] 
                    s_kmer = read2[-1] + 'X'
                else:
                    f_kmer = read1[position-1 : position+2]
                    s_kmer = read2[position-1] + 'X' + read2[position]    
            errorType = first + '-' + second
            new_line_list = [read1, read1_fre, line[2], errorType, position, f_kmer, s_kmer, read2, read2_fre, line[5]]
        else:
            raise ValueError("The edit distance of two reads in the input list must equal to one!")
        return new_line_list

    def real_ed1_seqs(self, total_seqs, read):
        """
        given a read, generate all its 1nt-edit-distance read counterparts existing in the dataset to form the edges  

        Args:
            total_seqs (list): The list consisting of all reads in the sequencing dataset.
            read (str): A DNA/RNA sequence.

        Returns:
            list: list of tuples of read pairs with only one base different
        """
        possible_ed1 = enumerate_ed1_seqs(read)
        real_seqs =  total_seqs.intersection(possible_ed1)
        read_lst = [read] * len(real_seqs)
        edges = list(zip(read_lst, real_seqs))
        return edges

    def extract_err_samples(self, graph, edit_dis):
        """
        extract genuine errors, error-free records, ambiguous errors or high ambiguous errors from read graph

        Args:
            graph (object): A graph constructed using NetworkX.
            edit_dis (int): set edit distance 1 or 2 to search edges for constructing graph

        Returns:
            DataFrame: three or four pandas dataframes saving genuine, error-free records, ambiguous errors or high ambiguous errors
        """
        # subgraphs = [graph.subgraph(c).copy() for c in nx.connected_components(graph)]
        # genuine errors and ambiguous errors
        # genuine_df, ambiguous_df = self.extract_genuine_ambi_errs(subgraphs, edit_dis)
        # genuine_df, ambiguous_df = self.extract_genuine_ambi_errs(graph, edit_dis)
        # isolated negative samples
        negative_df = self.extract_isolated_negatives(graph, edit_dis)
        # ambiguous errors
        if edit_dis == 1 and self.config.high_ambiguous: #
            # high_ambiguous_df = self.extract_high_ambiguous_errs(subgraphs)
            genuine_df, ambiguous_df, high_ambiguous_df = self.extract_genuine_ambi_errs(graph, edit_dis)
            return genuine_df, negative_df, ambiguous_df, high_ambiguous_df
        elif edit_dis == 1 or edit_dis == 2:
            genuine_df, ambiguous_df = self.extract_genuine_ambi_errs(graph, edit_dis)
            return genuine_df, negative_df, ambiguous_df

    def extract_umi_genuine_errs(self, input_file):  
        """
        extract genuine errors from umi graph

        Returns:
            DataFrame: one pandas dataframe saving genuine errors
        """
        graph, seqs_lens_lst, seqs2id_dict, unique_seqs = self.generate_graph(input_file, edit_dis=1)
        del seqs_lens_lst, seqs2id_dict, unique_seqs
        subgraphs = [graph.subgraph(c).copy() for c in nx.connected_components(graph) if len(c) >= 2 ]#
        # print(len(subgraphs))
        genuine_lst = []
        chunk_size = len(subgraphs) // self.config.chunks_num
        if chunk_size >= self.config.num_workers:
            groups = [subgraphs[i:i+chunk_size] for i in range(0, len(subgraphs), chunk_size)]

            # Write each group of subgraphs to separate files
            gexf_files = []
            # print(len(groups))
            for i, group in enumerate(groups):
                # Create a new graph for the group of subgraphs
                group_G = nx.Graph()
                # print(group)
                for subgraph_nodes in group:
                    # print(subgraph_nodes)
                    group_G.add_edges_from(graph.subgraph(subgraph_nodes).edges())
                    # Add node attributes to the new graph
                    for node in subgraph_nodes:
                        # print(node)
                        group_G.nodes[node].update(graph.nodes[node])
                # Generate the file name for the group
                file_name = self.config.result_dir + f"group_{i}.gexf"

                # Write the group of subgraphs to the file
                nx.write_gexf(group_G, file_name)
                gexf_files.append(file_name)
            del groups, subgraphs, graph

            for gexf_file in gexf_files:
                cur_graph = nx.read_gexf(gexf_file)
                sub_graphs = [cur_graph.subgraph(c).copy() for c in nx.connected_components(cur_graph)]
                
                subgraph_num = len(sub_graphs)
                try:
                    with WorkerPool(self.config.num_workers, shared_objects=sub_graphs, start_method='fork') as pool:
                        cur_genuine_lsts = pool.imap(self.extract_umi_genuine_errs_subgraph, range(subgraph_num))
                except KeyboardInterrupt:
                    # Handle termination signal (Ctrl+C)
                    pool.terminate()  # Terminate the WorkerPool before exiting
                except Exception:
                    # Handle other exceptions
                    pool.terminate()  # Terminate the WorkerPool before exiting
                    raise

                for item in cur_genuine_lsts:
                    genuine_lst.extend(item)

                os.remove(gexf_file)
                del cur_graph, sub_graphs
        else:
            subgraph_num = len(subgraphs)
            try:
                with WorkerPool(self.config.num_workers, shared_objects=subgraphs, start_method='fork') as pool:
                    cur_genuine_lsts = pool.imap(self.extract_umi_genuine_errs_subgraph, range(subgraph_num))
            except KeyboardInterrupt:
                # Handle termination signal (Ctrl+C)
                pool.terminate()  # Terminate the WorkerPool before exiting
            except Exception:
                # Handle other exceptions
                pool.terminate()  # Terminate the WorkerPool before exiting
                raise

            del subgraphs, graph

            for item in cur_genuine_lsts:
                genuine_lst.extend(item)

            
        genuine_df = pd.DataFrame(genuine_lst, columns=["StartRead","StartReadCount", "StartDegree", "ErrorTye","ErrorPosition", "StartErrKmer", "EndErrKmer", "EndRead", "EndReadCount", "EndDegree"])

        if self.config.verbose:
            genuine_csv = os.path.join(self.config.result_dir, "umi_gnuine.csv")
            genuine_df.to_csv(genuine_csv, index=False) 
        return genuine_df

    def extract_umi_genuine_errs_subgraph(self, sub_graphs, i): 
        gen_lst = []
        sub_graph = sub_graphs[i]
        nodes_lst = list(sub_graph.nodes)
        for node in nodes_lst:
            node_count = sub_graph.nodes[node]['count']
            node_degree = sub_graph.degree[node]
            line = []
            if node_degree >= 1 and node_count <= self.config.max_error_freq and not sub_graph.nodes[node]['flag']: #and node_degree <= 4
                node_neis = [n for n in sub_graph.neighbors(node)]
                # nei2count = []
                nei_degree_count = []
                for nei in node_neis:
                    nei_count = sub_graph.nodes[nei]['count']
                    nei_degree = sub_graph.degree[nei]
                    # nei2count.append((nei, nei_count))
                    tt = nei_degree * 0.5 + nei_count * 0.5
                    nei_degree_count.append((nei, tt, nei_count))
                # nei2count.sort(key=lambda x:x[1], reverse=True)
                nei_degree_count.sort(key=lambda x:x[1], reverse=True)
                # first_nei, first_nei_count = nei2count[0]
                first_nei, tt, first_nei_count = nei_degree_count[0]
                first_nei_degree = sub_graph.degree[first_nei]
                if first_nei_count > self.config.high_freq_thre:
                    line = [first_nei, first_nei_count, first_nei_degree, node, node_count, node_degree]
                    newline = self.err_type_classification(line)
                    # genuine_df.loc[len(genuine_df)] = newline
                    gen_lst.append(newline)
                    sub_graph.nodes[node]['flag'] = True
            else:
                continue   
        return gen_lst
        
    def extract_genuine_ambi_errs(self, graph, edit_dis):
        """
        extract genuine and ambiguous errors from read graph

        Args:
            subgraphs (class): Subgraphs of graph constructed using NetworkX.
            edit_dis (int): set edit distance 1 or 2 to search edges for constructing graph

        Returns:
            DataFrame: two pandas dataframes saving genuine and ambiguous errors
        """
        if edit_dis == 1:
            genu_columns = ["StartRead","StartReadCount", "StartDegree", "ErrorTye","ErrorPosition", "StartErrKmer", "EndErrKmer", "EndRead", "EndReadCount", "EndDegree"]
            ambiguous_df= pd.DataFrame(columns=["idx", "StartRead","StartReadCount", "StartDegree", "ErrorTye","ErrorPosition", "StartErrKmer", "EndErrKmer", "EndRead", "EndReadCount", "EndDegree"])
            genuine_csv = self.config.result_dir + "genuine1.csv"
            ambiguous_csv = self.config.result_dir + "ambiguous1.csv"
        elif edit_dis == 2: #or edit_dis == 3:
            genu_columns = ["StartRead","StartReadCount", "StartDegree", "EndRead", "EndReadCount", "EndDegree"]
            ambiguous_df= pd.DataFrame(columns=["idx", "StartRead", "StartReadCount", "StartDegree", "EndRead", "EndReadCount", "EndDegree"])
            genuine_csv = self.config.result_dir + "genuine2.csv"
            ambiguous_csv = self.config.result_dir + "ambiguous2.csv"  

        subgraphs = [graph.subgraph(c).copy() for c in nx.connected_components(graph) if len(c) >= 2 ]#
        # subgraphs = [graph.subgraph(c).copy() for c in nx.connected_components(graph)]

        self.logger.info("Extracting genuine and ambiguous errors...")
        genuine_lst = []
        ambiguous_lst = []

        if self.config.high_ambiguous:
            high_ambiguous_df = pd.DataFrame(columns=["idx", "StartRead", "StartReadCount", "StartDegree", "ErrorTye","ErrorPosition", "StartErrKmer", "EndErrKmer", "EndRead", "EndReadCount", "EndDegree"])    
            high_ambi_lst = []  
            high_idx = 0

        chunk_size = len(subgraphs) // self.config.chunks_num
        # chunk_size = self.config.num_workers
        # chunk_num = len(subgraphs) // chunk_size
        self.logger.debug(chunk_size)
        if chunk_size >= self.config.num_workers:
        # if chunk_num > 1:
        # if False:
            groups = [subgraphs[i:i+chunk_size] for i in range(0, len(subgraphs), chunk_size)]
            # self.logger.debug(groups)
            # Write each group of subgraphs to separate files
            gexf_files = []
            for i, group in enumerate(groups):
                # self.logger.debug(group)
                # Create a new graph for the group of subgraphs
                group_G = nx.Graph()
                for subgraph_nodes in group:
                    group_G.add_edges_from(graph.subgraph(subgraph_nodes).edges())
                    # Add node attributes to the new graph

                    for node in subgraph_nodes:
                        # self.logger.debug(node)
                        group_G.nodes[node].update(graph.nodes[node])
                # Generate the file name for the group
                file_name = self.config.result_dir + f"group_{i}.gexf"
                # Write the group of subgraphs to the file
                nx.write_gexf(group_G, file_name)
                gexf_files.append(file_name)
            del groups, subgraphs, graph

            for gexf_file in gexf_files:
                cur_graph = nx.read_gexf(gexf_file)
                sub_graphs = [cur_graph.subgraph(c).copy() for c in nx.connected_components(cur_graph)]
                try:
                    subgraph_num = len(sub_graphs)
                    shared_obs = sub_graphs, edit_dis
                    with WorkerPool(self.config.num_workers, shared_objects=shared_obs, start_method='fork') as pool:
                        for genu_ambi_lst in pool.imap(self.extract_genuine_ambi_errs_subgraph, range(subgraph_num)): # progress_bar=self.config.verbose
                            if genu_ambi_lst[0]:
                                genuine_lst.extend(genu_ambi_lst[0])
                                ambiguous_lst.extend(genu_ambi_lst[1])
                    del shared_obs
                except KeyboardInterrupt:
                    # Handle termination signal (Ctrl+C)
                    pool.terminate()  # Terminate the WorkerPool before exiting
                except Exception:
                    # Handle other exceptions
                    pool.terminate()  # Terminate the WorkerPool before exiting
                    raise

                if self.config.high_ambiguous:
                    cur_lst, cur_idx = self.extract_high_ambiguous_errs(sub_graphs, high_idx)
                    high_ambi_lst.extend(cur_lst)
                    high_idx = cur_idx + 1
                os.remove(gexf_file)
                del cur_graph, sub_graphs
        else:    
            try:
                subgraph_num = len(subgraphs)
                shared_obs = subgraphs, edit_dis
                with WorkerPool(self.config.num_workers, shared_objects=shared_obs, start_method='fork') as pool:
                    for genu_ambi_lst in pool.imap(self.extract_genuine_ambi_errs_subgraph, range(subgraph_num)): # progress_bar=self.config.verbose
                        if genu_ambi_lst[0]:
                            genuine_lst.extend(genu_ambi_lst[0])
                            ambiguous_lst.extend(genu_ambi_lst[1])
                del shared_obs
            except KeyboardInterrupt:
                # Handle termination signal (Ctrl+C)
                pool.terminate()  # Terminate the WorkerPool before exiting
            except Exception:
                # Handle other exceptions
                pool.terminate()  # Terminate the WorkerPool before exiting
                raise
            if self.config.high_ambiguous:
                cur_lst, cur_idx = self.extract_high_ambiguous_errs(subgraphs, high_idx)
                high_ambi_lst.extend(cur_lst)
                high_idx = cur_idx + 1

            del subgraphs, graph

        genuine_df = pd.DataFrame(genuine_lst, columns=genu_columns)

        idx = 0
        for a_item in ambiguous_lst:
            for sub_item in a_item:
                sub_item.insert(0, idx)
                ambiguous_df.loc[len(ambiguous_df)] = sub_item
            idx += 1

        if self.config.high_ambiguous: 
            for item in high_ambi_lst:
                high_ambiguous_df.loc[len(high_ambiguous_df)] = item
            
        if self.config.verbose:
            genuine_df.to_csv(genuine_csv, index=False)  
            ambiguous_df.to_csv(ambiguous_csv, index=False) 
            if self.config.high_ambiguous:   
                high_ambiguous_csv = self.config.result_dir + "high_ambiguous_1nt.csv"
                high_ambiguous_df.to_csv(high_ambiguous_csv, index=False)  
        self.logger.info("Done!")
        if edit_dis == 1 and self.config.high_ambiguous:
            return genuine_df, ambiguous_df, high_ambiguous_df
        elif edit_dis == 1 or edit_dis == 2:
            return genuine_df, ambiguous_df
    
    def extract_simplify_genuine_ambi_errs(self, graph, edit_dis):
        """
        extract genuine and ambiguous errors from read graph

        Args:
            subgraphs (class): Subgraphs of graph constructed using NetworkX.
            edit_dis (int): set edit distance 1 or 2 to search edges for constructing graph

        Returns:
            DataFrame: two pandas dataframes saving genuine and ambiguous errors
        """
        if edit_dis == 1:
            genu_columns = ["StartRead","StartReadCount", "StartDegree", "ErrorTye","ErrorPosition", "StartErrKmer", "EndErrKmer", "EndRead", "EndReadCount", "EndDegree"]
            ambiguous_df= pd.DataFrame(columns=["idx", "StartRead","StartReadCount", "StartDegree", "ErrorTye","ErrorPosition", "StartErrKmer", "EndErrKmer", "EndRead", "EndReadCount", "EndDegree"])
            genuine_csv = self.config.result_dir + "genuine1.csv"
            ambiguous_csv = self.config.result_dir + "ambiguous1.csv"
        elif edit_dis == 2: #or edit_dis == 3:
            genu_columns = ["StartRead","StartReadCount", "StartDegree", "EndRead", "EndReadCount", "EndDegree"]
            ambiguous_df= pd.DataFrame(columns=["idx", "StartRead", "StartReadCount", "StartDegree", "EndRead", "EndReadCount", "EndDegree"])
            genuine_csv = self.config.result_dir + "genuine2.csv"
            ambiguous_csv = self.config.result_dir + "ambiguous2.csv"  

        subgraphs = [graph.subgraph(c).copy() for c in nx.connected_components(graph) if len(c) >= 2 ]#
        # subgraphs = [graph.subgraph(c).copy() for c in nx.connected_components(graph)]
        # print("test")
        # print(len(subgraphs))
        
        self.logger.info("Extracting genuine and ambiguous errors...")
        genuine_lst = []
        ambiguous_lst = []
        # chunk_size = len(subgraphs) // self.config.chunks_num
        # chunk_size = self.config.num_workers
        # chunk_num = len(subgraphs) // chunk_size
        # if chunk_size >= self.config.num_workers:
        # if chunk_num > 1:
        if False:
            groups = [subgraphs[i:i+chunk_size] for i in range(0, len(subgraphs), chunk_size)]

            self.logger.debug(len(groups))
            # Write each group of subgraphs to separate files
            gexf_files = []
            for i, group in enumerate(groups):
                # Create a new graph for the group of subgraphs
                # self.logger.debug(group)
                group_G = nx.Graph()
                for subgraph_nodes in group:
                    group_G.add_edges_from(graph.subgraph(subgraph_nodes).edges())
                    # Add node attributes to the new graph
                    # self.logger.debug(subgraph_nodes)
                    for node in subgraph_nodes:
                        # self.logger.debug(node)
                        group_G.nodes[node].update(graph.nodes[node])
                # Generate the file name for the group
                file_name = self.config.result_dir + f"group_{i}.gexf"

                # Write the group of subgraphs to the file
                nx.write_gexf(group_G, file_name)
                gexf_files.append(file_name)
            del groups, subgraphs, graph
            ####
            for gexf_file in gexf_files:
                cur_graph = nx.read_gexf(gexf_file)
                sub_graphs = [cur_graph.subgraph(c).copy() for c in nx.connected_components(cur_graph)]
                try:
                    subgraph_num = len(sub_graphs)
                    shared_obs = sub_graphs, edit_dis
                    with WorkerPool(self.config.num_workers, shared_objects=shared_obs, start_method='fork') as pool:
                        for genu_ambi_lst in pool.imap(self.extract_genuine_ambi_errs_subgraph, range(subgraph_num)): # progress_bar=self.config.verbose
                            if genu_ambi_lst[0]:
                                genuine_lst.extend(genu_ambi_lst[0])
                                ambiguous_lst.extend(genu_ambi_lst[1])
                    del shared_obs
                except KeyboardInterrupt:
                    # Handle termination signal (Ctrl+C)
                    pool.terminate()  # Terminate the WorkerPool before exiting
                except Exception:
                    # Handle other exceptions
                    pool.terminate()  # Terminate the WorkerPool before exiting
                    raise
                os.remove(gexf_file)
                del cur_graph, sub_graphs
        # else:
        try:
            subgraph_num = len(subgraphs)
            shared_obs = subgraphs, edit_dis
            with WorkerPool(self.config.num_workers, shared_objects=shared_obs, start_method='fork') as pool:
                for genu_ambi_lst in pool.imap(self.extract_genuine_ambi_errs_subgraph, range(subgraph_num)): # progress_bar=self.config.verbose
                    if genu_ambi_lst[0]:
                        genuine_lst.extend(genu_ambi_lst[0])
                        ambiguous_lst.extend(genu_ambi_lst[1])
            del shared_obs
        except KeyboardInterrupt:
            # Handle termination signal (Ctrl+C)
            pool.terminate()  # Terminate the WorkerPool before exiting
        except Exception:
            # Handle other exceptions
            pool.terminate()  # Terminate the WorkerPool before exiting
            raise
        del subgraphs, graph

        genuine_df = pd.DataFrame(genuine_lst, columns=genu_columns)

        idx = 0
        for a_item in ambiguous_lst:
            for sub_item in a_item:
                sub_item.insert(0, idx)
                ambiguous_df.loc[len(ambiguous_df)] = sub_item
            idx += 1

        if self.config.verbose:
            genuine_df.to_csv(genuine_csv, index=False)  
            ambiguous_df.to_csv(ambiguous_csv, index=False) 
        self.logger.info("Done!")
        return genuine_df, ambiguous_df


    def add_genu_sample(self, shared_obs, i):
        """
        add samples to pd DataFrame

        Args:
            item_lst (list): a list containing samples

        Returns:
            DataFrame: pd dataframe saving one sample one row
        """
        item_lst, edit_dis = shared_obs
        if edit_dis == 1:
            tmp_df = pd.DataFrame(columns=["StartRead","StartReadCount", "StartDegree", "ErrorTye","ErrorPosition", "StartErrKmer", "EndErrKmer", "EndRead", "EndReadCount", "EndDegree"])
        elif edit_dis == 2:
            tmp_df = pd.DataFrame(columns=["StartRead","StartReadCount", "StartDegree", "EndRead", "EndReadCount", "EndDegree"])
        # for item in item_lst:
        tmp_df.loc[len(tmp_df)] = item_lst[i]
        return tmp_df

    def add_ambi_sample(self, idx, item_lst):
        """
        add ambiguous error samples to pd DataFrame

        Args:
            item_lst (list): a list containing samples

        Returns:
            DataFrame: pd dataframe saving one sample one row
        """
        tmp_df = pd.DataFrame()
        for a_item in item_lst:
            for sub_item in a_item:
                sub_item.insert(0, idx)
                tmp_df.loc[len(tmp_df)] = sub_item
            idx += 1
        return tmp_df

    '''
    def extract_genuine_ambi_errs_subgraph(self, sub_graph, edit_dis):
        gen_lst = []
        ambi_lst = []
        nodes_lst = list(sub_graph.nodes)
        for node in nodes_lst:
            node_count = sub_graph.nodes[node]['count']
            node_degree = sub_graph.degree[node]
            if node_count <= self.config.max_error_freq and not sub_graph.nodes[node]['flag']:
                node_neis = [n for n in sub_graph.neighbors(node)]
                if node_degree == 1:
                    nei = node_neis[0]
                    nei_count = sub_graph.nodes[nei]['count']
                    nei_degree = sub_graph.degree[nei]
                    # if nei_count >= self.config.high_freq_thre:
                    if nei_count > self.config.high_freq_thre:
                        line = [nei, nei_count, nei_degree, node, node_count, node_degree]
                        if edit_dis == 1:
                            newline = self.err_type_classification(line)
                            # gen_df.loc[len(gen_df)] = newline
                            gen_lst.append(newline)
                            del line, newline 
                        else:
                            # gen_df.loc[len(gen_df)] = line 
                            gen_lst.append(line)
                            del line                                   
                # elif node_degree <= self.config.ambiguous_error_node_degree: # comment this line on 13 May 2023
                else:
                    high_num = 0
                    nei2count = []
                    for nei in node_neis:
                        nei_count = sub_graph.nodes[nei]['count']
                        nei_degree = sub_graph.degree[nei]
                        # if nei_count >= self.config.high_freq_thre:
                        if nei_count > self.config.high_freq_thre:
                            high_num += 1
                            nei2count.append((nei, nei_count))
                    if high_num == 1:             
                        nei2count.sort(key=lambda x:x[1], reverse=True)
                        first_nei, first_nei_count = nei2count[0]
                        first_nei_degree = sub_graph.degree[first_nei]
                        # if first_nei_count >= self.config.high_freq_thre:
                        if first_nei_count > self.config.high_freq_thre:
                            line = [first_nei, first_nei_count, first_nei_degree, node, node_count, node_degree]
                            if edit_dis == 1:
                                newline = self.err_type_classification(line)
                                # gen_df.loc[len(gen_df)] = newline
                                gen_lst.append(newline)
                                del line, newline  
                            else:
                                # gen_df.loc[len(gen_df)] = line 
                                gen_lst.append(line)
                                del line
                    else:
                        # ambiguous errors
                        tmp_lst = []
                        for cre_nei, cur_nei_count in nei2count:
                            # if cur_nei_count >= self.config.high_freq_thre:
                            if cur_nei_count > self.config.high_freq_thre:
                                cur_nei_degree = sub_graph.degree[nei]
                                line = [cre_nei, cur_nei_count, cur_nei_degree, node, node_count, node_degree]
                                if edit_dis == 1:
                                    newline = self.err_type_classification(line)
                                else:
                                    newline = line
                                # newline.insert(0, idx)
                                # ambi_df.loc[len(ambi_df)] = newline
                                tmp_lst.append(newline)
                                del newline 
                        if tmp_lst:
                            ambi_lst.append(tmp_lst)  
                        # idx += 1 
                sub_graph.nodes[node]['flag'] = True        
            else:
                continue
        return gen_lst, ambi_lst
    '''
    
    def data_files(self, edit_dis):
        """
        function to return the results produced by DataGneration class

        Args:
            edit_dis (int): set edit distance 1 or 2 to search edges for constructing graph

        Returns:
            MultiVariables: MultiVariables for next step error correction
        """
        # 1nt-edit-distance-based graph
        # self.logger.info("-------------------------------------------------------------")
        # self.logger.info("1nt-edit-distance read graph error correction")
        self.MM.start()
        graph, seqs_lens_lst, seqs2id_dict, unique_seqs = self.generate_graph(self.config.input_file, edit_dis)
        seq_max_len = max(seqs_lens_lst)
        seq_min_len = min(seqs_lens_lst)
        self.logger.debug(seqs_lens_lst)
        self.logger.debug("Reads Max length: {}".format(seq_max_len))
        self.logger.debug("Reads Min length: {}".format(seq_min_len))
        self.MM.measure()
        if edit_dis == 1 and self.config.high_ambiguous:
            genuine_df, negative_df, ambiguous_df, high_ambiguous_df = self.extract_err_samples(graph, edit_dis)
        elif edit_dis == 2 or edit_dis == 1:
            genuine_df, negative_df, ambiguous_df = self.extract_err_samples(graph, edit_dis)
        self.MM.measure()

        isolates_file, non_isolates_file = self.extract_isolates(graph, unique_seqs, seqs2id_dict)
        self.MM.measure()
        self.MM.stop()
        del graph
        if self.config.high_ambiguous:
            return isolates_file, non_isolates_file, unique_seqs, seq_max_len, seq_min_len, genuine_df, negative_df, ambiguous_df, high_ambiguous_df
        else:
            return isolates_file, non_isolates_file, unique_seqs, seq_max_len, seq_min_len, genuine_df, negative_df, ambiguous_df

    def simplify_data_files(self, input_f, edit_dis):
        """
        function to return the results produced by DataGneration class

        Args:
            edit_dis (int): set edit distance 1 or 2 to search edges for constructing graph

        Returns:
            MultiVariables: MultiVariables for next step error correction
        """
        self.MM.start()
        # 1nt-edit-distance-based graph
        # self.logger.info("-------------------------------------------------------------")
        # self.logger.info("1nt-edit-distance read graph error correction")
        if edit_dis == 1:
            graph, seqs_lens_lst, seqs2id_dict, unique_seqs = self.generate_graph(input_f, edit_dis)
            seq_max_len = max(seqs_lens_lst)
            seq_min_len = min(seqs_lens_lst)
            self.logger.debug(seqs_lens_lst)
            self.logger.debug("Reads Max length: {}".format(seq_max_len))
            self.logger.debug("Reads Min length: {}".format(seq_min_len))
            self.MM.measure()
            genuine_df, ambiguous_df = self.extract_simplify_genuine_ambi_errs(graph, edit_dis)
            self.MM.measure()
            isolates_file, non_isolates_file = self.extract_isolates(graph, unique_seqs, seqs2id_dict)
            del graph
            self.MM.measure()
            self.MM.stop()
            return isolates_file, non_isolates_file, seq_max_len, seq_min_len, genuine_df, ambiguous_df
        elif edit_dis == 2:
            self.logger.debug(input_f)
            # if self.seq_min_len > 30:   
            self.logger.info("Constructing 2nt-edit-distance based graph.")
            graph, unique_seqs = self.generate_graph(input_f, edit_dis=2)
            self.graph_summary(graph)  
            self.MM.measure()
            genuine_df, ambiguous_df = self.extract_simplify_genuine_ambi_errs(graph, edit_dis)
            self.MM.measure()
            self.MM.stop()
            return genuine_df, ambiguous_df       

    def extract_isolated_negatives(self, graph, edit_dis):
        """
        extract isolated high frequency reads from read graph

        Args:
            graph (object): A graph constructed using NetworkX.
            edit_dis (int): set edit distance 1 or 2 to search edges for constructing graph

        Returns:
            DataFrame: one pandas dataframe saving isolated high frequency reads
        """
        self.logger.info("Extracting negative samples...")
        if not nx.is_connected(graph):
            self.logger.debug("G is a connected graph: {}".format(nx.is_connected(graph)))
            isolates = set(list(nx.isolates(graph)))        
        else:
            self.logger.debug("G is a connected graph: {}".format(nx.is_connected(graph)) )

        self.logger.debug("Isolated Nodes Number: {}".format(len(isolates)))
        # save isolated nodes as negative samples
        if edit_dis == 1:
            negative_csv = self.config.result_dir + "negative1.csv"
        elif edit_dis == 2:   
            negative_csv = self.config.result_dir + "negative2.csv"  
        negative_df = pd.DataFrame(columns=["StartRead","StartReadCount", "StartDegree"])
        for k in isolates:
            k_count = graph.nodes[k]['count']
            if k_count > self.config.high_freq_thre:
                k_degree = graph.degree[k]
                line = [k, k_count, k_degree]
                negative_df.loc[len(negative_df)] = line  
        if self.config.verbose:
            negative_df.to_csv(negative_csv, index=False) 
        self.logger.info("Done!")
        return negative_df

    def extract_isolates(self, graph, unique_seqs, seqs2id_dict):
        """
        split the source file into two files of isolates and non-isolates based on constructed graph

        Args:
            graph (object): A graph constructed using NetworkX.
            unique_seqs (list): All the unique reads in the dataset
            seqs2id_dict (dict): key: read, id: sequencing id

        Returns:
            files: two files of isolates and non-isolates based on constructed graph
        """
        self.logger.info("Extracting isolated nodes...")
        if not nx.is_connected(graph):
            self.logger.debug("G is a connected graph: {}".format(nx.is_connected(graph)))
            isolates = set(list(nx.isolates(graph)))
        else:
            self.logger.debug("G is a connected graph: {}".format(nx.is_connected(graph)) )

        self.logger.debug("Isolated Nodes Number: {}".format(len(isolates)))

        name_lst = []
        non_name_lst = []
        nonisolated_seqs = unique_seqs - isolates 
        # save isolated nodes as negative samples
        for k in isolates:
            name_lst.extend(seqs2id_dict[k])                  
                
        for s in nonisolated_seqs:
            non_name_lst.extend(seqs2id_dict[s])

        self.logger.debug("isolates name list: {}".format(len(name_lst)))
        self.logger.debug("non-isolates name list: {}".format(len(non_name_lst)))

        bases = self.config.input_file.split('/')[-1]
        base = bases.split('.')

        if self.file_type == 'fastq' or self.file_type == 'fq' or self.file_type == 'fastq.gz' or self.file_type == 'fq.gz':
            isolates_file = self.config.result_dir + base[0] + '_isolates.fastq'
            non_isolates_file = self.config.result_dir + base[0] + '_non_isolates.fastq'
        elif self.file_type == 'fasta' or self.file_type == 'fa' or self.file_type == 'fasta.gz' or self.file_type == 'fas.gz':
            isolates_file = self.config.result_dir + base[0] + '_isolates.fasta'
            non_isolates_file = self.config.result_dir + base[0] + '_non_isolates.fasta' 

        extract_records(self.config.result_dir, name_lst, self.config.input_file, isolates_file)
        extract_records(self.config.result_dir, non_name_lst, self.config.input_file, non_isolates_file)

        self.logger.info("Done!")
        return isolates_file, non_isolates_file

    def generate_graph(self, data_set, edit_dis):
        """
        construct 1nt- or 2nt-edit-distance-based read graph

        Args:
            data_set (str): The filename including path to be corrected.
            edit_dis (int): set edit distance 1 or 2 to search edges for constructing graph

        Returns:
            MultiVariables: Multi Variables after constructing edit-distance-based read graph
        """
        # if not os.path.exists(data_set):
        #     self.logger.error("No input file!")
        #     # os._exit(0)
        # else:
        self.logger.info("Input dataset '% s'" % data_set)
        record_iterator, file_type = parse_data(data_set)
        seqs2id_dict = {}
        total_seqs = []
        seq_lens_set = set()
        for item in record_iterator:
            seq = str(item.seq)
            ll = len(seq)
            seq_lens_set.add(ll)
            total_seqs.append(seq)
            seqs2id_dict.setdefault(seq, []).append(str(item.id))
        unique_seqs = set(total_seqs)

        self.logger.info("Constructing " + str(edit_dis) + "nt-edit-distance read graph...")
        graph = nx.Graph()
        read_count = Counter(total_seqs)
        high_freq = []
        low_freq = []

        for read, frequency in read_count.items():
            if not graph.has_node(read):
                graph.add_node(read, count = frequency, flag=False)  
            # if frequency >= self.config.high_freq_thre:
            if frequency > self.config.high_freq_thre:
                high_freq.append(read)
            else:
                low_freq.append(read)
        if len(high_freq) == 0:
            self.logger.error("Error Correction Failed as no high-frequency reads detected.")
            sys.exit(1)
        self.logger.debug(len(read_count))

        ######################################################
        edges_lst = []
        if edit_dis == 1:
            shared_unique_seqs = unique_seqs
        elif edit_dis == 2:
            shared_unique_seqs = low_freq
        
        self.logger.debug("Searching edges for constructing " + str(edit_dis) + "nt-edit-distance read graph...")
        try:
            with WorkerPool(self.config.num_workers, shared_objects=shared_unique_seqs, start_method='fork') as pool:
                if edit_dis == 1:
                    for edge_lst in pool.imap(self.real_ed1_seqs, high_freq, progress_bar=self.config.verbose):
                        edges_lst.extend(edge_lst)
                elif edit_dis == 2:
                    for edge_lst in pool.imap(self.real_ed2_seqs, high_freq, progress_bar=self.config.verbose):
                        edges_lst.extend(edge_lst)
        except KeyboardInterrupt:
            # Handle termination signal (Ctrl+C)
            pool.terminate()  # Terminate the WorkerPool before exiting
        except Exception:
            # Handle other exceptions
            pool.terminate()  # Terminate the WorkerPool before exiting
            raise
        del shared_unique_seqs
        if len(edges_lst) > 0:
            self.logger.debug(len(edges_lst))
            self.logger.debug(edges_lst[0])
            graph.add_edges_from(edges_lst)
        # self.logger.info(str(edit_dis) + "nt-edit-distance read graph construction done.")
        self.logger.info("Done!")
        ########################################################
        # save graphs
        if self.config.graph_visualization or self.config.save_graph:
            if edit_dis == 1:
                subdir = self.config.result_dir + "graph1/"                
            elif edit_dis == 2:
                subdir = self.config.result_dir + "graph2/"
            if os.path.exists(subdir):
                self.logger.info("Directory '%s' already exists" % subdir)
            else:
                os.makedirs(subdir)
                self.logger.info("Directory '%s' created" % subdir)
        if self.config.save_graph:
            graph_filename = subdir + 'graph.gexf'
            nx.write_gexf(graph, graph_filename)
            self.logger.info("Graph file *.gexf saved!")
        if self.config.graph_visualization:
            self.draw_graph(graph, subdir, self.config.drawing_graph_num)
            self.logger.info("Graph file *.png saved!")
        if edit_dis == 1:
            return graph, seq_lens_set, seqs2id_dict, unique_seqs
        else:
            return graph, unique_seqs

    def real_ed2_seqs(self, low_seqs, read):
        """
        given a read, generate all its 2nt-edit-distance substitution counterparts existing in the dataset to form the edges

        Args:
            low_seqs (list): The list consisting of all reads with low-frequency in the sequencing dataset.
            read (str): A DNA/RNA sequence.

        Returns:
            list: list of tuples of read pairs with two bases different
        """
        # possible_ed2 = set(self.ed2_seqs(read))
        possible_ed2 = enumerate_ed2_seqs(read)
        real_ed2_seqs = possible_ed2.intersection(low_seqs)
        # real_ed2_seqs = []
        # for seq in possible_ed2:
        #     if seq in low_seqs:
        #         real_ed2_seqs.append(seq)
        read_lst = [read] * len(real_ed2_seqs)
        edges = list(zip(read_lst, real_ed2_seqs))     
        return edges   

    def extract_ed2_errors(self, data_set): 
        """
        extract genuine errors, error-free records, ambiguous errors and unique reads from 2nt-edit-distance-based read graph

        Args:
            data_set (str): The filename including path for 2nt errors correction.

        Returns:
            DataFrame: three pandas dataframes of genuine_df, negative_df and ambiguous_df
            list: unique_seqs save unique reads of input dataset
        """
        self.logger.debug(data_set)
        self.MM.start()
        # if self.seq_min_len > 30:   
        self.logger.info("Constructing 2nt-edit-distance based graph.")
        graph, unique_seqs = self.generate_graph(data_set, edit_dis=2)
        self.MM.measure()
        self.graph_summary(graph)
        genuine_df, negative_df, ambiguous_df = self.extract_err_samples(graph, edit_dis=2)
        del graph
        self.MM.measure()
        self.MM.stop()
        return genuine_df, negative_df, ambiguous_df, unique_seqs

    def extract_amplicon_err_samples(self, data_set):
        """
        extract amplicon sequencing errors from read graph

        Args:
            data_set (str): The filename including path for further errors correction for amplicon sequencing data.

        Returns:
            DataFrame: amplicon_df, a dataframe saving the amplicon sequencing errors
            List: unique_seqs, a list saving unique reads of input dataset
        """
        self.MM.start()
        self.logger.debug(data_set)
        self.logger.info("Constructing the second 1nt-edit-distance based graph for further amplicon sequencing data correction.")
        graph, seq_lens_set, seqs2id_dict, unique_seqs = self.generate_graph(data_set, edit_dis=1)
        self.graph_summary(graph)
        self.MM.measure()
        del seq_lens_set, seqs2id_dict, unique_seqs
        subgraphs = [graph.subgraph(c).copy() for c in nx.connected_components(graph) if len(c) >= 2]
        
        amplicon_df = pd.DataFrame(columns=["idx", "StartRead","StartReadCount", "StartDegree", "ErrorTye", "ErrorPosition", "StartErrKmer", "EndErrKmer", "EndRead", "EndReadCount", "EndDegree"])
        for sub_graph in subgraphs:
            edges_lst = [e for e in sub_graph.edges()]
            if len(edges_lst) > 0:
                nodes_lst = list(sub_graph.nodes)
                for node in nodes_lst:
                    # reset node visit status
                    sub_graph.nodes[node]['flag'] = False

        chunk_size = len(subgraphs) // self.config.chunks_num
        idx = 0
        amplicon_lst = []

        if chunk_size >= self.config.num_workers:
            groups = [subgraphs[i:i+chunk_size] for i in range(0, len(subgraphs), chunk_size)]
            # Write each group of subgraphs to separate files
            gexf_files = []
            for i, group in enumerate(groups):
                # Create a new graph for the group of subgraphs
                group_G = nx.Graph()
                for subgraph_nodes in group:
                    group_G.add_edges_from(graph.subgraph(subgraph_nodes).edges())
                    # Add node attributes to the new graph
                    for node in subgraph_nodes:
                        group_G.nodes[node].update(graph.nodes[node])
                # Generate the file name for the group
                file_name = self.config.result_dir + f"group_{i}.gexf"

                # Write the group of subgraphs to the file
                nx.write_gexf(group_G, file_name)
                gexf_files.append(file_name)
            del groups, subgraphs, graph

            for gexf_file in gexf_files:
                cur_graph = nx.read_gexf(gexf_file)
                sub_graphs = [cur_graph.subgraph(c).copy() for c in nx.connected_components(cur_graph)]
                cur_lst, cur_idx = self.extract_amplicon_errs(sub_graphs, idx)
                amplicon_lst.extend(cur_lst)
                idx = cur_idx + 1
                os.remove(gexf_file)
                del cur_graph, sub_graphs
        else:
            cur_lst, cur_idx = self.extract_amplicon_errs(subgraphs, idx)
            amplicon_lst.extend(cur_lst)
            idx = cur_idx + 1   
            del subgraphs, graph  

        for item in amplicon_lst:
            amplicon_df.loc[len(amplicon_df)] = item
        if self.config.verbose:
            amplicon_df.to_csv(self.config.result_dir + "amplicon.csv", index=False) 
        self.MM.measure()
        self.MM.stop() 
        return amplicon_df
        
    def extract_amplicon_errs(self, subgraphs, idx):  
        amplicon_errs_lst = []     
        for sub_graph in subgraphs:
            edges_lst = [e for e in sub_graph.edges()]
            if len(edges_lst) > 0:
                nodes_lst = list(sub_graph.nodes)
                for node in nodes_lst:
                    node_count = sub_graph.nodes[node]['count']
                    node_degree = sub_graph.degree[node]
                    line = []
                    if node_count < self.config.amplicon_low_freq and not sub_graph.nodes[node]['flag']:# and node_degree <= self.config.amplicon_error_node_degree:
                        node_neis = [n for n in sub_graph.neighbors(node)]
                        # nei2count = []
                        nei_degree_count = []
                        for nei in node_neis:
                            nei_count = sub_graph.nodes[nei]['count']
                            nei_degree = sub_graph.degree[nei]
                            if nei_count > self.config.amplicon_high_freq:
                                line = [nei, nei_count, nei_degree, node, node_count, node_degree]
                                new_line = self.err_type_classification(line) 
                                new_line.insert(0, idx) 
                                # writer.writerow(new_line)
                                # amplicon_df.loc[len(amplicon_df)] = new_line
                                amplicon_errs_lst.append(new_line)
                        idx += 1
                        sub_graph.nodes[node]['flag'] = True
                    else:
                        continue     
        return amplicon_errs_lst, idx
  
    def draw_graph(self, graph, sub_dir, drawing_graph_num = 50):
        """
        draw subgraphs from edit-distance-based read graph

        Args:
            graph (object): A graph constructed using NetworkX.
            sub_dir (str): Directory for saving subgraphs.
            drawing_graph_num (int, optional): The number of subgraphs to be drawed. Defaults to 50.
        """
        
        # layout graphs with positions using graphviz neato
        # pos = nx.nx_agraph.graphviz_layout(graph, prog="neato")
        # color nodes the same in each connected subgraph
        i = 0
        S = [graph.subgraph(c).copy() for c in nx.connected_components(graph)]
        for s in S:
            if nx.number_of_nodes(s) >= 10:
                labels = nx.get_node_attributes(s, 'count')
                # A = nx.nx_agraph.to_agraph(s)
                # print(type(A))
                # A.layout(prog="sfdp", args="-Nshape=point")
                # A.draw(graph_name + str(i) + ".png",  args="")
                nodes_num = s.number_of_nodes()
                fig_size = int(((nodes_num // 100)) * 1.5) + 6
                plt.figure(1, figsize=(fig_size, fig_size))
                nx.draw(s, pos = graphviz_layout(s, prog='sfdp', args="-Nshape=circle"), labels=labels, font_color="black", node_color="#fbb4ae", edge_color="#66c2a5")
                plt.savefig(sub_dir + str(i) + "_" + str(nodes_num) +  ".svg", format="svg")
                i += 1
                del nodes_num
                plt.close()
                if i == drawing_graph_num:
                    break
        return

    def extract_high_ambiguous_errs(self, subgraphs, idx):
        """
        extract high ambiguous errors from read graph

        Args:
            subgraphs (class): Subgraphs of graph constructed using NetworkX.

        Returns:
            DataFrame: One pandas dataframe saving high ambiguous errors
        """
        high_ambi_lst = []
        for s in subgraphs:
            edges_lst = [e for e in s.edges()]
            for (a, b) in edges_lst:
                a_count = s.nodes[a]['count']
                b_count = s.nodes[b]['count']
                a_degree = s.degree[a]
                b_degree = s.degree[b]
                if a_count > self.config.high_freq_thre and b_count > self.config.high_freq_thre: 
                    a2b = [a, a_count, a_degree, b, b_count, b_degree]
                    new_a2b = self.err_type_classification(a2b) 
                    new_a2b.insert(0, idx)     
                    high_ambi_lst.append(new_a2b)
                    # high_ambiguous_df.loc[len(high_ambiguous_df)] = new_a2b 
                    b2a = [b, b_count, b_degree, a, a_count, a_degree]
                    new_b2a = self.err_type_classification(b2a)
                    new_b2a.insert(0, idx)
                    high_ambi_lst.append(new_b2a)
                    # high_ambiguous_df.loc[len(high_ambiguous_df)] = new_b2a
                    idx += 1
            del edges_lst
        # if self.config.verbose:
        #     high_ambiguous_csv = self.config.result_dir + "high_ambiguous_1nt.csv"
        #     high_ambiguous_df.to_csv(high_ambiguous_csv, index=False)  
        # self.logger.info("Done!")
        # return high_ambiguous_df
        return high_ambi_lst, idx


    '''
    def extract_genuine_ambi_errs(self, subgraphs, edit_dis):
        """
        extract genuine and ambiguous errors from read graph

        Args:
            subgraphs (class): Subgraphs of graph constructed using NetworkX.
            edit_dis (int): set edit distance 1 or 2 to search edges for constructing graph

        Returns:
            DataFrame: two pandas dataframes saving genuine and ambiguous errors
        """
        if edit_dis == 1:
            genuine_df = pd.DataFrame(columns=["StartRead","StartReadCount", "StartDegree", "ErrorTye","ErrorPosition", "StartErrKmer", "EndErrKmer", "EndRead", "EndReadCount", "EndDegree"])
            ambiguous_df= pd.DataFrame(columns=["idx", "StartRead","StartReadCount", "StartDegree", "ErrorTye","ErrorPosition", "StartErrKmer", "EndErrKmer", "EndRead", "EndReadCount", "EndDegree"])
            genuine_csv = self.config.result_dir + "genuine1.csv"
            ambiguous_csv = self.config.result_dir + "ambiguous1.csv"
        elif edit_dis == 2 or edit_dis == 3:
            genuine_df = pd.DataFrame(columns=["StartRead","StartReadCount", "StartDegree", "EndRead", "EndReadCount", "EndDegree"])
            ambiguous_df= pd.DataFrame(columns=["idx", "StartRead", "StartReadCount", "StartDegree", "EndRead", "EndReadCount", "EndDegree"])
            genuine_csv = self.config.result_dir + "genuine2.csv"
            ambiguous_csv = self.config.result_dir + "ambiguous2.csv"  
        idx = 0
        for sub_graph in tqdm(subgraphs, desc=self.logger.info("Extract samples with genuine errors"), miniters=int(len(subgraphs)/self.config.min_iters)):
            edges_lst = [e for e in sub_graph.edges()]
            if len(edges_lst) > 0:
                nodes_lst = list(sub_graph.nodes)
                for node in nodes_lst:
                    node_count = sub_graph.nodes[node]['count']
                    node_degree = sub_graph.degree[node]
                    if node_count <= self.config.max_error_freq and not sub_graph.nodes[node]['flag']:
                        node_neis = [n for n in sub_graph.neighbors(node)]
                        if node_degree == 1:
                            nei = node_neis[0]
                            nei_count = sub_graph.nodes[nei]['count']
                            nei_degree = sub_graph.degree[nei]
                            if nei_count >= self.config.high_freq_thre:
                                line = [nei, nei_count, nei_degree, node, node_count, node_degree]
                                if edit_dis == 1:
                                    newline = self.err_type_classification(line)
                                    genuine_df.loc[len(genuine_df)] = newline
                                    del line, newline 
                                else:
                                    genuine_df.loc[len(genuine_df)] = line 
                                    del line                                   
                        else:
                            high_num = 0
                            nei2count = []
                            for nei in node_neis:
                                nei_count = sub_graph.nodes[nei]['count']
                                nei_degree = sub_graph.degree[nei]
                                if nei_count >= self.config.high_freq_thre:
                                    high_num += 1
                                    nei2count.append((nei, nei_count))
                            if high_num == 1:             
                                nei2count.sort(key=lambda x:x[1], reverse=True)
                                first_nei, first_nei_count = nei2count[0]
                                first_nei_degree = sub_graph.degree[first_nei]
                                if first_nei_count >= self.config.high_freq_thre:
                                    line = [first_nei, first_nei_count, first_nei_degree, node, node_count, node_degree]
                                    if edit_dis == 1:
                                        newline = self.err_type_classification(line)
                                        genuine_df.loc[len(genuine_df)] = newline
                                        del line, newline  
                                    else:
                                        genuine_df.loc[len(genuine_df)] = line 
                                        del line
                            else:
                                # ambiguous errors
                                for cre_nei, cur_nei_count in nei2count:
                                    if cur_nei_count >= self.config.high_freq_thre:
                                        cur_nei_degree = sub_graph.degree[nei]
                                        line = [cre_nei, cur_nei_count, cur_nei_degree, node, node_count, node_degree]
                                        if edit_dis == 1:
                                            newline = self.err_type_classification(line)
                                        else:
                                            newline = line
                                        newline.insert(0, idx)
                                        ambiguous_df.loc[len(ambiguous_df)] = newline
                                        del newline   
                                idx += 1 
                        sub_graph.nodes[node]['flag'] = True        
                    else:
                        continue
        if self.config.verbose:
            genuine_df.to_csv(genuine_csv, index=False)  
            ambiguous_df.to_csv(ambiguous_csv, index=False)    
        return genuine_df, ambiguous_df
    '''

    def extract_genuine_ambi_errs_subgraph(self, shared_obs, ii):
        gen_lst = []
        ambi_lst = []
        sub_graphs, edit_dis = shared_obs 
        sub_graph = sub_graphs[ii]
        edges_lst = [e for e in sub_graph.edges()]
        if len(edges_lst) > 0:
            nodes_lst = list(sub_graph.nodes)
            for node in nodes_lst:
                node_count = sub_graph.nodes[node]['count']
                node_degree = sub_graph.degree[node]
                if node_count <= self.config.max_error_freq and not sub_graph.nodes[node]['flag']:
                    node_neis = [n for n in sub_graph.neighbors(node)]
                    if node_degree == 1:
                        nei = node_neis[0]
                        nei_count = sub_graph.nodes[nei]['count']
                        nei_degree = sub_graph.degree[nei]
                        # if nei_count >= self.config.high_freq_thre:
                        if nei_count > self.config.high_freq_thre:
                            line = [nei, nei_count, nei_degree, node, node_count, node_degree]
                            if edit_dis == 1:
                                newline = self.err_type_classification(line)
                                # gen_df.loc[len(gen_df)] = newline
                                gen_lst.append(newline)
                                del line, newline 
                            else:
                                # gen_df.loc[len(gen_df)] = line 
                                gen_lst.append(line)
                                del line                                   
                    # elif node_degree <= self.config.ambiguous_error_node_degree: # comment this line on 13 May 2023
                    else:
                        high_num = 0
                        nei2count = []
                        for nei in node_neis:
                            nei_count = sub_graph.nodes[nei]['count']
                            nei_degree = sub_graph.degree[nei]
                            # if nei_count >= self.config.high_freq_thre:
                            if nei_count > self.config.high_freq_thre:
                                high_num += 1
                                nei2count.append((nei, nei_count))
                        if high_num == 1:             
                            nei2count.sort(key=lambda x:x[1], reverse=True)
                            first_nei, first_nei_count = nei2count[0]
                            first_nei_degree = sub_graph.degree[first_nei]
                            # if first_nei_count >= self.config.high_freq_thre:
                            if first_nei_count > self.config.high_freq_thre:
                                line = [first_nei, first_nei_count, first_nei_degree, node, node_count, node_degree]
                                if edit_dis == 1:
                                    newline = self.err_type_classification(line)
                                    # gen_df.loc[len(gen_df)] = newline
                                    gen_lst.append(newline)
                                    del line, newline  
                                else:
                                    # gen_df.loc[len(gen_df)] = line 
                                    gen_lst.append(line)
                                    del line
                        else:
                            # ambiguous errors
                            tmp_lst = []
                            for cre_nei, cur_nei_count in nei2count:
                                # if cur_nei_count >= self.config.high_freq_thre:
                                if cur_nei_count > self.config.high_freq_thre:
                                    cur_nei_degree = sub_graph.degree[nei]
                                    line = [cre_nei, cur_nei_count, cur_nei_degree, node, node_count, node_degree]
                                    if edit_dis == 1:
                                        newline = self.err_type_classification(line)
                                    else:
                                        newline = line
                                    # newline.insert(0, idx)
                                    # ambi_df.loc[len(ambi_df)] = newline
                                    tmp_lst.append(newline)
                                    del newline 
                            if tmp_lst:
                                ambi_lst.append(tmp_lst)  
                            # idx += 1 
                    sub_graph.nodes[node]['flag'] = True        
                else:
                    continue
        return [gen_lst, ambi_lst]


    '''
    def extract_high_ambiguous_errs(self, subgraphs):
        """
        extract high ambiguous errors from read graph

        Args:
            subgraphs (class): Subgraphs of graph constructed using NetworkX.

        Returns:
            DataFrame: One pandas dataframe saving high ambiguous errors
        """
        high_ambiguous_df = pd.DataFrame(columns=["idx", "StartRead", "StartReadCount", "StartDegree", "ErrorTye","ErrorPosition", "StartErrKmer", "EndErrKmer", "EndRead", "EndReadCount", "EndDegree"])
        idx = 0
        for s in tqdm(subgraphs, desc=self.logger.info("Extracting high ambiguous errors from 1nt-edit-distance graph"), miniters=int(len(subgraphs)/self.config.min_iters)):
            edges_lst = [e for e in s.edges()]
            if len(edges_lst) > 0:
                for (a, b) in edges_lst:
                    a_count = s.nodes[a]['count']
                    b_count = s.nodes[b]['count']
                    a_degree = s.degree[a]
                    b_degree = s.degree[b]
                    if a_count > self.config.high_freq_thre and b_count > self.config.high_freq_thre: 
                    # if a_count >= self.config.high_freq_thre and b_count >= self.config.high_freq_thre: # comment this line on 13 May 2023
                        a2b = [a, a_count, a_degree, b, b_count, b_degree]
                        new_a2b = self.err_type_classification(a2b) 
                        new_a2b.insert(0, idx)     
                        high_ambiguous_df.loc[len(high_ambiguous_df)] = new_a2b 
                        b2a = [b, b_count, b_degree, a, a_count, a_degree]
                        new_b2a = self.err_type_classification(b2a)
                        new_b2a.insert(0, idx)
                        high_ambiguous_df.loc[len(high_ambiguous_df)] = new_b2a
                        idx += 1
            del edges_lst
        if self.config.verbose:
            high_ambiguous_csv = self.config.result_dir + "high_ambiguous_1nt.csv"
            high_ambiguous_df.to_csv(high_ambiguous_csv, index=False)  
        self.logger.info("Done!")
        return high_ambiguous_df
    '''